--	SMart Heal
--	Written by scott@mithrandir.com
--	Scott's attempt to do something initially based on DruidPartyHeal and Healomatic
--	but with inspiration and code derived from many.

--	Notes
--
--	This does not use mass healing spells, only spells that target individuals.  E.g., for
--	Druids this does not use Tranquility, but only uses Healing Touch, Regrowth and Rejuvenation.
--
--	There is no way to determine LOS on the healee.  We can determine whether the healee is in range
--	but if not in LOS we will get an error message when attempting to cast.
--
--	There is no API that detects movement.  Therefore, we attempt an approximation using the
--	GetPlayerMapPosition() call to record X, Y coords for the character to guess when moving.
--	This system does not work well, so we sometimes attempt to cast something other than a
--	spell that is instant (like Rejuvenation).
--
--	TODO
--
--	Allow casting of buffs to target, self and party members (maybe even pets).  Should probably
--	have a preference to indicate that if not enough mana, do not cast at all.
--

SMH_NAME = "SMart Heal";
SMH_VERSION = "0.9.7";
SMH_VERSION_STRING = SMH_NAME.." "..SMH_VERSION;
SMH_MACRO_COMMAND = "/smh";

local _SMH_class;				-- class of the player.  used to determine whether we should enable the addon.
local _SMH_myName;				-- name of the player.  only use in determining raid group.
local _SMH_event_dispatch = {};			-- table of functions whose keys are the events
local _SMH_lastX = nil;				-- where we last were in X
local _SMH_lastY = nil;				-- where we last were in Y
local _SMH_isMoving = false;			-- determined using above two variables
local _SMH_healingSlot = nil;			-- indicates the slot in action bar that has one of the healing spells in which we have interest -- for range detection
local _SMH_swiftNatureSlot = nil;		-- indicates the slot in action bar that has Nature's Swiftness -- so we can check for cooldown for automatic casting
local _SMH_powerWordShieldSlot = nil;		-- indicates the slot in action bar that has Power Word: Shield -- so we can check for cooldown
local _SMH_extraHealingAmount = 0;		-- amount of extra healing that is done by equipment
local _SMH_maxNumberInColumn = 10;		-- maximum number of character names in each priority column
local _SMH_maxNumberInAlgColumn = 10;		-- maximum number of spells in the algorithm column
local _SMH_maxNumberInProColumn = 10;		-- maximum number of profiles in the profile column
local _SMH_currentTab = 0;			-- indicates current tab for use for refresh when profile changes
local _SMH_algorithmAddingSpell;		-- used to keep track of spell being added to algorithm

SMH_CLASS_DRUID_ENGLISH = 'DRUID';		-- the non-localized version used to determine if we are a Druid
SMH_CLASS_PALADIN_ENGLISH = 'PALADIN';
SMH_CLASS_PRIEST_ENGLISH = 'PRIEST';
SMH_CLASS_SHAMAN_ENGLISH = 'SHAMAN';

local classesPermittedToUseAddon = {
	[SMH_CLASS_DRUID_ENGLISH] = true,
	[SMH_CLASS_PALADIN_ENGLISH] = true,
	[SMH_CLASS_PRIEST_ENGLISH] = true,
	[SMH_CLASS_SHAMAN_ENGLISH] = true,
	};

local classSpellsOfInterest = {
	[SMH_CLASS_DRUID_ENGLISH] = { SMH_SPELL_HEALING_TOUCH, SMH_SPELL_REGROWTH, SMH_SPELL_REJUVENATION },
	[SMH_CLASS_PALADIN_ENGLISH] = { SMH_SPELL_FLASH_OF_LIGHT, SMH_SPELL_HOLY_LIGHT },
	[SMH_CLASS_PRIEST_ENGLISH] = { SMH_SPELL_FLASH_HEAL, SMH_SPELL_GREATER_HEAL, SMH_SPELL_HEAL, SMH_SPELL_LESSER_HEAL, SMH_SPELL_RENEW, SMH_SPELL_POWER_WORD_SHIELD },
	[SMH_CLASS_SHAMAN_ENGLISH] = { SMH_SPELL_HEALING_WAVE, SMH_SPELL_LESSER_HEALING_WAVE },
	};

--	Buffs in which we have interest are those that can be cast on other people and
--	have differing ranks, limited to those that work on a single individual at a
--	time.  All others are left to the manual operation of the caster.  Got to have
--	something to do, right?
local classBuffsOfInterest = {
	[SMH_CLASS_DRUID_ENGLISH] = { SMH_SPELL_MARK_OF_THE_WILD, SMH_SPELL_THORNS },
	[SMH_CLASS_PALADIN_ENGLISH] = { },
	[SMH_CLASS_PRIEST_ENGLISH] = { },
	[SMH_CLASS_SHAMAN_ENGLISH] = { },
	};

local classAttacksOfInterest = {
	[SMH_CLASS_DRUID_ENGLISH] = { SMH_SPELL_MOONFIRE },
	[SMH_CLASS_PALADIN_ENGLISH] = { },
	[SMH_CLASS_PRIEST_ENGLISH] = { },
	[SMH_CLASS_SHAMAN_ENGLISH] = { },
	};

local totems = {
	[SMH_TOTEM_ANCIENT_MANA_SPRING] = 1, [SMH_TOTEM_ATALAI] = 1, [SMH_TOTEM_BRAIN_WASH] = 1, [SMH_TOTEM_CORRUPTED] = 1, [SMH_TOTEM_DIRE_MAUL_CRYSTAL] = 1,
	[SMH_TOTEM_DISEASE_CLEANSING] = 1, [SMH_TOTEM_EARTHBIND] = 1, [SMH_TOTEM_EARTHGRAB] = 1, [SMH_TOTEM_ELEMENTAL_PROTECTION] = 1, [SMH_TOTEM_FIRE_NOVA] = 1,
	[SMH_TOTEM_FIRE_RESISTANCE] = 1, [SMH_TOTEM_FLAME_BUFFET] = 1, [SMH_TOTEM_FLAMETONGUE] = 1, [SMH_TOTEM_FROST_RESISTANCE] = 1, [SMH_TOTEM_GRACE_OF_AIR] = 1,
	[SMH_TOTEM_GROUNDING] = 1, [SMH_TOTEM_HEALING_STREAM] = 1, [SMH_TOTEM_HEALING] = 1, [SMH_TOTEM_ICE] = 1, [SMH_TOTEM_LAVA_SPOUT] = 1, [SMH_TOTEM_MAGMA] = 1,
	[SMH_TOTEM_MANA_SPRING] = 1, [SMH_TOTEM_MANA_TIDE] = 1, [SMH_TOTEM_MOONFLARE] = 1, [SMH_TOTEM_NATURE_RESISTANCE] = 1, [SMH_TOTEM_POISON_CLEANSING] = 1,
	[SMH_TOTEM_POWERFUL_HEALING] = 1, [SMH_TOTEM_SCORCHING] = 1, [SMH_TOTEM_SEARING] = 1, [SMH_TOTEM_SENTRY] = 5, [SMH_TOTEM_SPIRIT] = 1, [SMH_TOTEM_STONECLAW] = 5,
	[SMH_TOTEM_STONESKIN] = 1, [SMH_TOTEM_STRENGTH_OF_EARTH] = 1, [SMH_TOTEM_SUPERIOR_HEALING] = 1, [SMH_TOTEM_TRANQUIL_AIR] = 1, [SMH_TOTEM_TREMOR] = 1,
	[SMH_TOTEM_WINDFURY] = 1, [SMH_TOTEM_WINDWALL] = 1,
	};

-- This table is populated from the above with entries of the spell names
-- and values of true once the class is known.  It is used as a convenience
-- when determining if what we have encountered is a spell of interest.
local spellsOfInterest = { };

--	Two variables are used.  The first is the set of defaults that are always present in the system.
--	These are the values that are used when the user resets the options to default status.  This is
--	also used to ensure the user's stored variables have the proper keys in them.  The second variable
--	are the actual option values that the user sets.  These are persistent between runs of the program
--	and work across versions of this addon because we check to ensure old keys are removed and new
--	keys are added after the stored variable is loaded.
SMH_Configuration_Defaults = {
	["messagesParty"] = true,		-- controls whether the spell cast messages appear in the party chat window
	["debug"] = false,			-- indicates whether we are in debug mode
	["targetPriority"] = true,		-- indicates whether the "target" has priority over all else
	["healPercentage"] = 50,		-- indicates the percentage of the possible heal range to use to compare to the target's health
	["healingPets"] = true,			-- indicates whether party pets are considered in the healing list
	["checkingNaturesSwiftness"] = true,	-- indicates whether we think about casting Nature's Swiftness automatically as part of heal process
	["threshold"] = 100,			-- indicates percentage of health target must be below to be considered for healing
	["petThreshold"] = 100,			-- indicates percentage of health pets must be below to be considered for healing
	["selfThreshold"] = 40,			-- indicates percentage of health self must be below to be forced to head of healing list
	["healingRaid"] = true,			-- indicates whether raid members are to be healed
	["healingRaidPets"] = true,		-- indicates whether raid pets are to be healed
	["partyOverRaid"] = true,		-- indicates whether the party has priority over the raid for healing
	["healsParty"] = true,			-- indicates whether parts members other than self are healed
	["minimapButtonAngle"] = 356,		-- indicates the angle around the minimap for the location of the button that toggles options UI
	["priorityRaidGroup1"] = false,		-- indicates members of raid group 1 are given priority
	["priorityRaidGroup2"] = false,		-- indicates members of raid group 2 are given priority
	["priorityRaidGroup3"] = false,		-- indicates members of raid group 3 are given priority
	["priorityRaidGroup4"] = false,		-- indicates members of raid group 4 are given priority
	["priorityRaidGroup5"] = false,		-- indicates members of raid group 5 are given priority
	["priorityRaidGroup6"] = false,		-- indicates members of raid group 6 are given priority
	["priorityRaidGroup7"] = false,		-- indicates members of raid group 7 are given priority
	["priorityRaidGroup8"] = false,		-- indicates members of raid group 8 are given priority
	["priorityRaidPets"] = false,		-- indicates that pets of groups that have priority get the same priority
	["classPriorities"] = {
		[SMH_CLASS_DRUID] = 2,
		[SMH_CLASS_HUNTER] = 2,
		[SMH_CLASS_MAGE] = 2,
		[SMH_CLASS_PALADIN] = 2,
		[SMH_CLASS_PRIEST] = 2,
		[SMH_CLASS_ROGUE] = 2,
		[SMH_CLASS_SHAMAN] = 2,
		[SMH_CLASS_WARLOCK] = 2,
		[SMH_CLASS_WARRIOR] = 2,
		};				-- indicates the priority that classes have
	["classPriorityOverParty"] = false,	-- indicates high priority classes have higher priority than party members
	["charactersPriorityHigh"] = { },	-- indicates characters that have the highest priority
	["charactersPriorityLow"] = { },	-- indicates characters that have low priority
	["charactersPriorityIgnored"] = { },	-- indicates characters that will not be healed
	["shouldDisplayProfile"] = false,	-- indicates whether the profile will be displayed in the mini profile window
	};

SMH_Configuration = {};

local outputBooleanMappings = { [false] = SMH_OUTPUT_OPTION_OFF, [true] = SMH_OUTPUT_OPTION_ON };

--	The cached spell book data is a two dimensional table with three rows representing
--	each of the spells and ten columns representing the ranks.  At each point in the table
--	is the index into the spell book for that spell.
local cachedSpellBookData = { };

--	Note that levelLearned cannot be used for determining whether a target can receive
--	a spell because Healing Touch can be cast on anyone of any level it seems.
local cachedSpellTargetLevels = {
	[SMH_SPELL_HEALING_TOUCH] = { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
	[SMH_SPELL_REGROWTH] = { 12, 18, 24, 30, 36, 42, 48, 54, 60 },
	[SMH_SPELL_REJUVENATION] = { 4, 10, 16, 22, 28, 34, 40, 46, 52, 58, 60 },
	[SMH_SPELL_HOLY_LIGHT] = { 1, 1, 1, 1, 1, 1, 1, 1, 1 },
	[SMH_SPELL_FLASH_OF_LIGHT] = { 1, 1, 1, 1, 1, 1 },
	[SMH_SPELL_LESSER_HEAL] = { 1, 1, 1 },
	[SMH_SPELL_HEAL] = { 1, 1, 1, 1 },
	[SMH_SPELL_GREATER_HEAL] = { 1, 1, 1, 1, 1 },
	[SMH_SPELL_RENEW] = { 8, 14, 20, 26, 32, 38, 44, 50, 56, 60 },
	[SMH_SPELL_FLASH_HEAL] = { 1, 1, 1, 1, 1, 1, 1 },
	[SMH_SPELL_POWER_WORD_SHIELD] = { 6, 12, 18, 24, 30, 36, 42, 48, 54, 60 },
	[SMH_SPELL_HEALING_WAVE] = { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
	[SMH_SPELL_LESSER_HEALING_WAVE] = { 1, 1, 1, 1, 1, 1 },

	[SMH_SPELL_MARK_OF_THE_WILD] = { 1, 10, 20, 30, 40, 50, 60 },
	[SMH_SPELL_THORNS] = { 6, 14, 24, 34, 44, 54 },
	};
local levelLearned = {
	[SMH_SPELL_HEALING_TOUCH] = { 1, 8, 14, 20, 26, 32, 38, 44, 50, 56, 60 },
	[SMH_SPELL_REGROWTH] = { 12, 18, 24, 30, 36, 42, 48, 54, 60 },
	[SMH_SPELL_REJUVENATION] = { 4, 10, 16, 22, 28, 34, 40, 46, 52, 58, 60 },
	[SMH_SPELL_HOLY_LIGHT] = { 1, 6, 14, 22, 30, 38, 46, 54, 60 },
	[SMH_SPELL_FLASH_OF_LIGHT] = { 20, 26, 34, 42, 50, 58 },
	[SMH_SPELL_LESSER_HEAL] = { 1, 4, 10 },
	[SMH_SPELL_HEAL] = { 16, 22, 28, 34 },
	[SMH_SPELL_GREATER_HEAL] = { 40, 46, 52, 58, 60 },
	[SMH_SPELL_RENEW] = { 8, 14, 20, 26, 32, 38, 44, 50, 56, 60 },
	[SMH_SPELL_FLASH_HEAL] = { 20, 26, 32, 38, 44, 50, 56 },
	[SMH_SPELL_POWER_WORD_SHIELD] = { 6, 12, 18, 24, 30, 36, 42, 48, 54, 60 },
	[SMH_SPELL_HEALING_WAVE] = { 1, 6, 12, 18, 24, 32, 40, 48, 56, 60 },
	[SMH_SPELL_LESSER_HEALING_WAVE] = { 20, 28, 36, 44, 52, 60 },
	};

--	This table provides the penalties for the +healing equipment based on the casting
--	time of the spell or its duration (depending on type of spell).  It is easier to
--	have this in a table vice keeping track of casting times for spells and computing.
local timePenalties = {
	[SMH_SPELL_HEALING_TOUCH] = { 58, 43, 29, 15, 0, 0, 0, 0, 0, 0, 0 },
	[SMH_SPELL_REGROWTH] = { 21, 21, 21, 21, 21, 21, 21, 21, 21 },
	[SMH_SPELL_REJUVENATION] = { 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20 },
	[SMH_SPELL_HOLY_LIGHT] = { 29, 29, 29, 29, 29, 29, 29, 29, 29 },
	[SMH_SPELL_FLASH_OF_LIGHT] = { 58, 58, 58, 58, 58, 58 },
	[SMH_SPELL_LESSER_HEAL] = { 43, 43, 28 },
	[SMH_SPELL_HEAL] = { 15, 0, 0, 0 },
	[SMH_SPELL_GREATER_HEAL] = { 0, 0, 0, 0, 0 },
	[SMH_SPELL_RENEW] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	[SMH_SPELL_FLASH_HEAL] = { 58, 58, 58, 58, 58, 58, 58 },
	[SMH_SPELL_POWER_WORD_SHIELD] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	[SMH_SPELL_HEALING_WAVE] = { 58, 43, 29, 15, 15, 15, 15, 15, 15, 15 },
	[SMH_SPELL_LESSER_HEALING_WAVE] = { 58, 58, 58, 58, 58, 58 },
	};

local equipmentSlots = { "Head", "Neck", "Shoulder", "Back", "Chest", "Shirt", "Tabard", "Wrist", "Hands", "Waist", "Legs", "Feet", "Finger0", "Finger1", "Trinket0", "Trinket1", "MainHand", "SecondaryHand", "Ranged" };

local _smh_defaultSpellIcons = {
	[SMH_SPELL_REGROWTH] = "Interface\\Icons\\Spell_Nature_ResistNature",
	[SMH_SPELL_REJUVENATION] = "Interface\\Icons\\Spell_Nature_Rejuvenation",
	[SMH_SPELL_HEALING_TOUCH] = "Interface\\Icons\\Spell_Nature_HealingTouch",
	[SMH_SPELL_HOLY_LIGHT] = "Interface\\Icons\\Spell_Holy_HolyBolt",
	[SMH_SPELL_FLASH_OF_LIGHT] = "Interface\\Icons\\Spell_Holy_FlashHeal",
	[SMH_SPELL_LESSER_HEAL] = "Interface\\Icons\\Spell_Holy_LesserHeal",
	[SMH_SPELL_HEAL] = "Interface\\Icons\\Spell_Holy_Heal",
	[SMH_SPELL_GREATER_HEAL] = "Interface\\Icons\\Spell_Holy_GreaterHeal",
	[SMH_SPELL_RENEW] = "Interface\\Icons\\Spell_Holy_Renew",
	[SMH_SPELL_FLASH_HEAL] = "Interface\\Icons\\Spell_Holy_FlashHeal",
	[SMH_SPELL_POWER_WORD_SHIELD] = "Interface\\Icons\\Spell_Holy_PowerWordShield",
	[SMH_SPELL_HEALING_WAVE] = "Interface\\Icons\\Spell_Nature_HealingWaveGreater",
	[SMH_SPELL_LESSER_HEALING_WAVE] = "Interface\\Icons\\Spell_Nature_HealingWaveLesser",
	};

--	The desired spell names are not localized since they are only used in debug messages.
--	Don't change the order of this table since references require the same order.
local _smh_HealingModeConfiguration = {
	{ ["nuke"] = { SMH_SPELL_REGROWTH, SMH_SPELL_REJUVENATION }, ["desire"] = "Healing Touch" },
	{ ["nuke"] = { SMH_SPELL_HEALING_TOUCH, SMH_SPELL_REJUVENATION }, ["desire"] = "Regrowth" },
	{ ["nuke"] = { SMH_SPELL_HEALING_TOUCH, SMH_SPELL_REGROWTH }, ["desire"] = "Rejuvenation" },
	{ ["nuke"] = { SMH_SPELL_FLASH_OF_LIGHT }, ["desire"] = "Holy Light" },
	{ ["nuke"] = { SMH_SPELL_HOLY_LIGHT }, ["desire"] = "Flash of Light" },
	{ ["nuke"] = { SMH_SPELL_HEAL, SMH_SPELL_GREATER_HEAL, SMH_SPELL_RENEW, SMH_SPELL_FLASH_HEAL, SMH_SPELL_POWER_WORD_SHIELD }, ["desire"] = "Lesser Heal" },
	{ ["nuke"] = { SMH_SPELL_LESSER_HEAL, SMH_SPELL_GREATER_HEAL, SMH_SPELL_RENEW, SMH_SPELL_FLASH_HEAL, SMH_SPELL_POWER_WORD_SHIELD }, ["desire"] = "Heal" },
	{ ["nuke"] = { SMH_SPELL_LESSER_HEAL, SMH_SPELL_HEAL, SMH_SPELL_RENEW, SMH_SPELL_FLASH_HEAL, SMH_SPELL_POWER_WORD_SHIELD }, ["desire"] = "Greater Heal" },
	{ ["nuke"] = { SMH_SPELL_LESSER_HEAL, SMH_SPELL_HEAL, SMH_SPELL_GREATER_HEAL, SMH_SPELL_FLASH_HEAL, SMH_SPELL_POWER_WORD_SHIELD }, ["desire"] = "Renew" },
	{ ["nuke"] = { SMH_SPELL_LESSER_HEAL, SMH_SPELL_HEAL, SMH_SPELL_GREATER_HEAL, SMH_SPELL_RENEW, SMH_SPELL_POWER_WORD_SHIELD }, ["desire"] = "Flash Heal" },
	{ ["nuke"] = { SMH_SPELL_LESSER_HEALING_WAVE }, ["desire"] = "Healing Wave" },
	{ ["nuke"] = { SMH_SPELL_HEALING_WAVE }, ["desire"] = "Lesser Healing Wave" },
	{ ["nuke"] = { SMH_SPELL_LESSER_HEAL, SMH_SPELL_HEAL, SMH_SPELL_GREATER_HEAL, SMH_SPELL_RENEW, SMH_SPELL_FLASH_HEAL }, ["desire"] = "Power Word: Shield" },
	};

local playerIds = { party1 = 1, party2 = 2, party3 = 3, party4 = 4 };
local petIds = { pet = 0, partypet1 = 1, partypet2 = 2, partypet3 = 3, partypet4 = 4 };

local function outputChatMessage(message)
	if DEFAULT_CHAT_FRAME then
		DEFAULT_CHAT_FRAME:AddMessage(message)
	end
end

local function outputDebugMessage(message)
	if SMH_Configuration.profiles[SMH_Configuration.currentProfile].debug then
		outputChatMessage(message)
	end
end

--	These functions and their use taken from Healomatic
local function parseInteger(text, index)
	local i = 1;
	for integer in string.gfind(text, '%d+') do
		if i == index then
			return tonumber(integer);
		end
		i = i + 1;
	end
end
local function piFirst(text)	return parseInteger(text, 1);	end
local function piFirstThird(text)	return parseInteger(text, 1) + parseInteger(text, 3);	end
local function piSecond(text)	return parseInteger(text, 2);	end
local function piSecondThird(text)	return parseInteger(text, 2) + parseInteger(text, 3);	end
--	Mapping of the various healing spells to parsers to get their amounts
local spellParsersMaxHeal = {
	[SMH_CLASS_DRUID_ENGLISH] = { [SMH_SPELL_HEALING_TOUCH] = piSecond, [SMH_SPELL_REJUVENATION] = piFirst, [SMH_SPELL_REGROWTH] = piSecondThird },
	[SMH_CLASS_PALADIN_ENGLISH] = { [SMH_SPELL_HOLY_LIGHT] = piSecond, [SMH_SPELL_FLASH_OF_LIGHT] = piSecond },
	[SMH_CLASS_PRIEST_ENGLISH] = { [SMH_SPELL_LESSER_HEAL] = piSecond, [SMH_SPELL_HEAL] = piSecond, [SMH_SPELL_GREATER_HEAL] = piSecond, [SMH_SPELL_RENEW] = piFirst, [SMH_SPELL_FLASH_HEAL] = piSecond },
	[SMH_CLASS_SHAMAN_ENGLISH] = { [SMH_SPELL_HEALING_WAVE] = piSecond, [SMH_SPELL_LESSER_HEALING_WAVE] = piSecond },
	};
local spellParsersMinHeal = {
	[SMH_CLASS_DRUID_ENGLISH] = { [SMH_SPELL_HEALING_TOUCH] = piFirst, [SMH_SPELL_REJUVENATION] = piFirst, [SMH_SPELL_REGROWTH] = piFirstThird, },
	[SMH_CLASS_PALADIN_ENGLISH] = { [SMH_SPELL_HOLY_LIGHT] = piFirst, [SMH_SPELL_FLASH_OF_LIGHT] = piFirst },
	[SMH_CLASS_PRIEST_ENGLISH] = { [SMH_SPELL_LESSER_HEAL] = piFirst, [SMH_SPELL_HEAL] = piFirst, [SMH_SPELL_GREATER_HEAL] = piFirst, [SMH_SPELL_RENEW] = piFirst, [SMH_SPELL_FLASH_HEAL] = piFirst },
	[SMH_CLASS_SHAMAN_ENGLISH] = { [SMH_SPELL_HEALING_WAVE] = piFirst, [SMH_SPELL_LESSER_HEALING_WAVE] = piFirst },
	};

local function UnitIsHealable(unit)
	if not UnitExists(unit) then return false end
	if not UnitIsFriend('player', unit) then return false end
	if     UnitIsEnemy(unit, 'player') then return false end
	if     UnitCanAttack(unit, 'player') then return false end
	if     UnitCanAttack('player', unit) then return false end
	if     unit ~= 'player' and not UnitCanCooperate('player', unit) then return false end
	if not UnitIsConnected(unit) then return false end
	if     UnitIsDead(unit) then return false end
	if     UnitIsGhost(unit) then return false end
	return true
end

local function healthCompare(unit1, unit2)
	return unit1.percent < unit2.percent;
end

local function greaterCompare(one, two)
	return one > two;
end

local function numericGreaterCompare(one, two)
	return tonumber(one) > tonumber(two);
end

local function lesserCompare(one, two)
	return one < two;
end

local function hasBuff(unitName, buffName, checkingDebuff)
	local retval = false;
	local texture;
	for i=0,15,1 do
		if checkingDebuff then texture = UnitDebuff(unitName, i) else texture = UnitBuff(unitName, i) end;
		if texture then
			if checkingDebuff then smhTooltip:SetUnitDebuff(unitName, i) else smhTooltip:SetUnitBuff(unitName, i) end;
			if smhTooltipTextLeft1:GetText() == buffName then
				retval = true;
				outputDebugMessage(unitName.." has buff "..buffName);
				break;
			end
		end
	end
	return retval;
end

--	This returns a table containing keys of the buff names.
local function unitBuffs(unitName)
	local retval = { };
	for i = 0, 15, 1 do
		texture = UnitBuff(unitName, i);
		if texture then
			smhTooltip:SetUnitBuff(unitName, i);
			retval[smhTooltipTextLeft1:GetText()] = true;
		end
		texture = UnitDebuff(unitName, i);
		if texture then
			smhTooltip:SetUnitDebuff(unitName, i);
			retval[smhTooltipTextLeft1:GetText()] = true;
		end
	end
	return retval;
end

local function findHealingSlot()
	outputDebugMessage("==> Finding healing slot");
	_SMH_healingSlot = nil;
	_SMH_swiftNatureSlot = nil;
	_SMH_powerWordShieldSlot = nil;

	for i = 1, 120 do
		if (HasAction(i)) then
			smhTooltip:SetAction(i);
			local slotName = smhTooltipTextLeft1:GetText();

			if spellsOfInterest[slotName] then
				_SMH_healingSlot = i;
			elseif _SMH_class == SMH_CLASS_DRUID_ENGLISH and slotName == SMH_SPELL_NATURESSWIFTNESS then
				_SMH_swiftNatureSlot = i;
			elseif _SMH_class == SMH_CLASS_PRIEST_ENGLISH and slotName == SMH_SPELL_POWER_WORD_SHIELD then
				_SMH_powerWordShieldSlot = i;
			end

		end
		if (_SMH_healingSlot ~= nil
			and (_SMH_swiftNatureSlot ~= nil or _SMH_class ~= SMH_CLASS_DRUID_ENGLISH)
			and (_SMH_powerWordShieldSlot ~= nil or _SMH_class ~= SMH_CLASS_PRIEST_ENGLISH)
			) then break end
	end
	if (_SMH_healingSlot ~= nil) then outputDebugMessage("Healing slot is ".._SMH_healingSlot); else outputDebugMessage("!!! Did Not find healing slot!!!"); end;
	if _SMH_class == SMH_CLASS_DRUID_ENGLISH then
		if _SMH_swiftNatureSlot ~= nil then outputDebugMessage("Nat. Swift. slot is ".. _SMH_swiftNatureSlot); else outputDebugMessage("!!! Did Not find Nat. Swift. slot!!!"); end;
	end
	if _SMH_class == SMH_CLASS_PRIEST_ENGLISH then
		if (_SMH_powerWordShieldSlot ~= nil) then outputDebugMessage("Power Word: Shield slot is ".. _SMH_powerWordShieldSlot) else outputDebugMessage("Did not find Power Word: Shield slot!!!") end;
	end
end

local function tryToInsertUnitForHealing(tableName, unitName, parameterName, debugMessage)
	local currentHealth = UnitHealth(unitName);
	local maxHealth = UnitHealthMax(unitName);
	local percentHealth = currentHealth / maxHealth;
	-- Only put damaged units that meet the damage threshold into the table.
	if (currentHealth < maxHealth and percentHealth < (SMH_Configuration.profiles[SMH_Configuration.currentProfile][parameterName] / 100.0)) then
		table.insert(tableName, 1, { id = unitName, percent = percentHealth, current = currentHealth, max = maxHealth });
		if (debugMessage ~= nil) then outputDebugMessage(debugMessage) end
	end
end

function SMH_priorityString()
	local retval = SMH_TIER_TITLE;
	local order = 2;
	local wordsToAdd;
	local tableSize;
	local priority = { [0] = { }, [1] = { }, [2] = { }, [3] = { } };

	-- We now take each of the class priorities and put them into the priority table.
	local temp = { };
	for n in pairs(SMH_Configuration.profiles[SMH_Configuration.currentProfile].classPriorities) do table.insert(temp, n) end
	table.sort(temp, greaterCompare);
	for i, n in ipairs(temp) do
		table.insert(priority[SMH_Configuration.profiles[SMH_Configuration.currentProfile].classPriorities[n]], n);
	end

	if SMH_Configuration.profiles[SMH_Configuration.currentProfile].targetPriority then
		retval = retval.."   "..order.. SMH_TIER_TARGET;
		order = order + 1;
	end

	retval = retval.."   "..order.. SMH_TIER_PLAYER1 ..SMH_Configuration.profiles[SMH_Configuration.currentProfile].selfThreshold.. SMH_TIER_PLAYER2;
	order = order + 1;

	temp = { };
	for n in pairs(SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityHigh) do table.insert(temp, n) end
	if table.getn(temp) > 0 then
		retval = retval .. "   " .. order .. ". ";
		table.sort(temp, lesserCompare);
		for i, n in ipairs(temp) do
			if i ~= 1 then retval = retval .. ", " end
			retval = retval .. n;
		end
		retval = retval .. "\n";
		order = order + 1;
	end

	if SMH_Configuration.profiles[SMH_Configuration.currentProfile].classPriorityOverParty and table.getn(priority[3]) > 0 then
		retval = retval .. "   " .. order .. ". ";
		for i = 1, table.getn(priority[3]), 1 do
			if i ~= 1 then
				retval = retval .. "/";
			end
			retval = retval .. priority[3][i];
		end
		retval = retval .. "\n";
		order = order + 1;
	end

	if SMH_Configuration.profiles[SMH_Configuration.currentProfile].partyOverRaid then
		wordsToAdd = { SMH_TIER_PLAYER };
		if SMH_Configuration.profiles[SMH_Configuration.currentProfile].healsParty then
			table.insert(wordsToAdd, SMH_TIER_PARTY);
		end
		if SMH_Configuration.profiles[SMH_Configuration.currentProfile].healingPets then
			table.insert(wordsToAdd, SMH_TIER_PARTYPETS);
		end
		tableSize = table.getn(wordsToAdd);
		if tableSize > 0 then
			retval = retval.."   "..order..". ";
			for i = 1, tableSize, 1 do
				if (i ~= 1) then
					retval = retval.."/";
				end
				retval = retval..wordsToAdd[i];
			end
			retval = retval.."\n";
			order = order + 1;
		end
	end

	if not SMH_Configuration.profiles[SMH_Configuration.currentProfile].classPriorityOverParty and table.getn(priority[3]) > 0 then
		local code = "";
		if SMH_Configuration.profiles[SMH_Configuration.currentProfile].partyOverRaid then code = SMH_TIER_RAIDCODE end
		retval = retval .. "   " .. order .. ". ";
		for i = 1, table.getn(priority[3]), 1 do
			if i ~= 1 then
				retval = retval .. "/";
			end
			retval = retval .. priority[3][i] .. code;
		end
		retval = retval .. "\n";
		order = order + 1;
	end

	wordsToAdd = { };
	for i = 1, 8, 1 do
		local key = "priorityRaidGroup"..i;
		if SMH_Configuration.profiles[SMH_Configuration.currentProfile][key] then
			table.insert(wordsToAdd, SMH_TIER_RAIDGROUP..i);
			if SMH_Configuration.profiles[SMH_Configuration.currentProfile].priorityRaidPets then
				table.insert(wordsToAdd, SMH_TIER_RAIDGROUP..i.. SMH_TIER_RAIDGROUPPETS);
			end
		end
	end
	tableSize = table.getn(wordsToAdd);
	if tableSize > 0 then
		retval = retval.."   "..order..". ";
		for i = 1, tableSize, 1 do
			if (i ~= 1) then
				retval = retval.."/";
			end
			retval = retval..wordsToAdd[i];
		end
		retval = retval.."\n";
		order = order + 1;
	end

	wordsToAdd = { };
	if SMH_Configuration.profiles[SMH_Configuration.currentProfile].healingRaid then
		table.insert(wordsToAdd, SMH_TIER_RAID);
	end
	if SMH_Configuration.profiles[SMH_Configuration.currentProfile].healingRaidPets then
		table.insert(wordsToAdd, SMH_TIER_RAIDPETS);
	end
	if not SMH_Configuration.profiles[SMH_Configuration.currentProfile].partyOverRaid then
		table.insert(wordsToAdd, SMH_TIER_PLAYER);
		if SMH_Configuration.profiles[SMH_Configuration.currentProfile].healsParty then
			table.insert(wordsToAdd, SMH_TIER_PARTY);
		end
		if SMH_Configuration.profiles[SMH_Configuration.currentProfile].healingPets then
			table.insert(wordsToAdd, SMH_TIER_PARTYPETS);
		end
	end
	tableSize = table.getn(wordsToAdd);
	if tableSize > 0 then
		retval = retval.."   "..order..". ";
		for i = 1, tableSize, 1 do
			if (i ~= 1) then
				retval = retval.."/";
			end
			retval = retval..wordsToAdd[i];
		end
		retval = retval.."\n";
		order = order + 1;
	end

	if table.getn(priority[1]) > 0 then
		retval = retval .. "   " .. order .. ". ";
		for i = 1, table.getn(priority[1]), 1 do
			if i ~= 1 then
				retval = retval .. "/";
			end
			retval = retval .. priority[1][i];
		end
		retval = retval .. "\n";
		order = order + 1;
	end

	temp = { };
	for n in pairs(SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityLow) do table.insert(temp, n) end
	if table.getn(temp) > 0 then
		retval = retval .. "   " .. order .. ". ";
		table.sort(temp, lesserCompare);
		for i, n in ipairs(temp) do
			if i ~= 1 then retval = retval .. ", " end
			retval = retval .. n;
		end
		retval = retval .. "\n";
		order = order + 1;
	end

	if table.getn(priority[0]) > 0 then
		retval = retval .. SMH_TIER_IGNORED_CLASSES;
		for i = 1, table.getn(priority[0]), 1 do
			if i ~= 1 then
				retval = retval .. ", ";
			end
			retval = retval .. priority[0][i];
		end
		retval = retval .. "\n";
	end

	temp = { };
	for n in pairs(SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityIgnored) do table.insert(temp, n) end
	if table.getn(temp) > 0 then
		retval = retval .. SMH_TIER_IGNORED_CHARACTERS;
		table.sort(temp, lesserCompare);
		for i, n in ipairs(temp) do
			if i ~= 1 then retval = retval .. ", " end
			retval = retval .. n;
		end
	end

	com_mithrandir_smhPriorityDesc:SetText(retval);
end

local function whichPriorityLevel(classModifier, isParty)
	local retval = classModifier + 1;

	if isParty then
		if classModifier == 3 then
			if SMH_Configuration.profiles[SMH_Configuration.currentProfile].classPriorityOverParty then
				retval = 7;
			elseif SMH_Configuration.profiles[SMH_Configuration.currentProfile].partyOverRaid then
				retval = 6;
			else
				retval = 5;
			end
		elseif classModifier == 2 then
			if SMH_Configuration.profiles[SMH_Configuration.currentProfile].partyOverRaid then
				retval = 6;
			end
		end
	else
		if classModifier == 3 then
			if SMH_Configuration.profiles[SMH_Configuration.currentProfile].classPriorityOverParty then
				retval = 7;
			else
				retval = 5;
			end
		end
	end

	return retval;
end

local function actionNotInCooldown(actionSlot)
	local retval = false;
	local start, duration, enable = GetActionCooldown(actionSlot);
	if start == 0 and duration == 0 then
		retval = true;
	end
	return retval;
end

local maxClassHealth = { [SMH_CLASS_DRUID] = 3000, [SMH_CLASS_HUNTER] = 3000, [SMH_CLASS_MAGE] = 2000, [SMH_CLASS_PALADIN] = 4000, [SMH_CLASS_PRIEST] = 2000, [SMH_CLASS_ROGUE] = 3000, [SMH_CLASS_SHAMAN] = 3000, [SMH_CLASS_WARLOCK] = 2000, [SMH_CLASS_WARRIOR] = 4000 };

local function orderedHealableUnits()
	local groupStatus = { };	-- the real results
	local tableSize;
	local tableBase;
	local classModifier;
	-- Note that we consider priority[8] higher than priority[1]
	local priority = { };		-- a table of tables
	local numberPriorities = 8;
	local characterName;

	-- initialize our priority tables to empty
	for i = 1, numberPriorities, 1 do
		priority[i] = { };
	end

	-- We now consider ourselves (as part of our own party)
	classModifier = SMH_Configuration.profiles[SMH_Configuration.currentProfile].classPriorities[UnitClass("player")];
	if classModifier ~= 0 then
		tableBase = whichPriorityLevel(classModifier, true);
		tryToInsertUnitForHealing(priority[tableBase], "player", "threshold", nil);
	end

	if SMH_Configuration.profiles[SMH_Configuration.currentProfile].healsParty then
		for i in playerIds do
			if UnitIsHealable(i) then
				classModifier = SMH_Configuration.profiles[SMH_Configuration.currentProfile].classPriorities[UnitClass(i)];
				characterName = UnitName(i);
				if classModifier ~= 0 and not SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityIgnored[characterName] then
					if SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityHigh[characterName] then tableBase = 8
					elseif SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityLow[characterName] then tableBase = 1
					else tableBase = whichPriorityLevel(classModifier, true)
					end
					tryToInsertUnitForHealing(priority[tableBase], i, "threshold", nil);
				end
			end
		end
	end

	if SMH_Configuration.profiles[SMH_Configuration.currentProfile].healingPets then
		for i in petIds do
			if UnitIsHealable(i) then
				characterName = UnitName(i);
				if not SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityIgnored[characterName] then
					if SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityHigh[characterName] then tableBase = 8
					elseif SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityLow[characterName] then tableBase = 1
					elseif SMH_Configuration.profiles[SMH_Configuration.currentProfile].partyOverRaid then tableBase = 6 else tableBase = 3 end
					tryToInsertUnitForHealing(priority[tableBase], i, "petThreshold", nil);
				end
			end
		end
	end

	-- Determine if any of the raid subgroups have priority
	local priorityRaid = false;
	for i = 1, 8, 1 do
		if SMH_Configuration.profiles[SMH_Configuration.currentProfile]["priorityRaidGroup"..i] then
			priorityRaid = true;
		end
	end

	-- Only process the raid code if we need to do so.
	-- Note that since we do not know what subgroup the player is in we
	-- first record each of the damaged ones into subgroups because we
	-- will ignore the player's subgroup if we can.
	if SMH_Configuration.profiles[SMH_Configuration.currentProfile].healingRaid or SMH_Configuration.profiles[SMH_Configuration.currentProfile].healingRaidPets or priorityRaid then
		local numberInRaid = GetNumRaidMembers();
		if (numberInRaid > 0) then
			local myRaidGroup = nil;
			local statuses = { };
			local petStatuses = { };
			local key;
			local targetTable;

			-- initialize the arrays to empty
			for i = 1, 8, 1 do
				key = "priorityRaidGroup"..i;
				statuses[key] = { };
				petStatuses[key] = { };
			end

			-- go through the entire raid to get the appropriate healees
			for i = 1, numberInRaid do
				local raidName, _, raidGroup = GetRaidRosterInfo(i);
				local unitName;

				key = "priorityRaidGroup"..raidGroup;

				if (raidName == _SMH_myName) then
					myRaidGroup = raidGroup;
				end

				if SMH_Configuration.profiles[SMH_Configuration.currentProfile].healingRaid or SMH_Configuration.profiles[SMH_Configuration.currentProfile][key] then
					unitName = "raid"..i;
					if UnitIsHealable(unitName) then
						tryToInsertUnitForHealing(statuses[key], unitName, "threshold", nil);
					end
				end
				if SMH_Configuration.profiles[SMH_Configuration.currentProfile].healingRaidPets or (SMH_Configuration.profiles[SMH_Configuration.currentProfile][key] and SMH_Configuration.profiles[SMH_Configuration.currentProfile].priorityRaidPets) then
					unitName = "raidpet"..i;
					if UnitIsHealable(unitName) then
						tryToInsertUnitForHealing(petStatuses[key], unitName, "petThreshold", nil);
					end
				end
			end

			-- now put the priority raid people into priority2Table and everyone else into groupStatus
			for i = 1, 8, 1 do
				local shouldProcess = true;
				key = "priorityRaidGroup"..i;

				if i == myRaidGroup then
					if not SMH_Configuration.profiles[SMH_Configuration.currentProfile][key] then
						shouldProcess = false;
					end
				end
				if shouldProcess then
					tableSize = table.getn(statuses[key]);
					for i = 1, tableSize do
						classModifier = SMH_Configuration.profiles[SMH_Configuration.currentProfile].classPriorities[UnitClass(statuses[key][i].id)];
						characterName = UnitName(statuses[key][i].id);
						if classModifier ~= 0 and not SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityIgnored[characterName] then
							if SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityHigh[characterName] then tableBase = 8
							elseif SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityLow[characterName] then tableBase = 1
							elseif SMH_Configuration.profiles[SMH_Configuration.currentProfile][key] then tableBase = 4 else tableBase = whichPriorityLevel(classModifier, false) end
							table.insert(priority[tableBase], 1, statuses[key][i]);
						end
					end

					if SMH_Configuration.profiles[SMH_Configuration.currentProfile][key] and SMH_Configuration.profiles[SMH_Configuration.currentProfile].priorityRaidPets then tableBase = 4 else tableBase = 3 end
					tableSize = table.getn(petStatuses[key]);
					for i = 1, tableSize do
						characterName = UnitName(petStatuses[key][i].id);
						if not SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityIgnored[characterName] then
							if SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityHigh[characterName] then tableBase = 8
							elseif SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityLow[characterName] then tableBase = 1
							end
							table.insert(priority[tableBase], 1, petStatuses[key][i]);
						end
					end
				end
			end
		end
	end

	-- sort each of the priority tables individually to maintain order within priority
	for i = 1, numberPriorities, 1 do
		if table.getn(priority[i]) > 0 then
			table.sort(priority[i], healthCompare);
		end
	end

	-- and then add them to the overall table
	for i = 1, numberPriorities, 1 do
		tableSize = table.getn(priority[i]);
		if tableSize > 0 then
			for j = 1, tableSize, 1 do
				table.insert(groupStatus, j, priority[i][j]);
			end
		end
	end

	--	Put myself at front of list if my health is less than 40%
	tryToInsertUnitForHealing(groupStatus, 'player', "selfThreshold", "Player put at head of heal list");

	--	Put the target at the front of the list if it is healable and damaged
	if SMH_Configuration.profiles[SMH_Configuration.currentProfile].targetPriority and UnitIsHealable('target') then
		local targetCurrentHealth = UnitHealth('target');
		local targetMaxHealth = UnitHealthMax('target');
		local shouldAddTarget = false;

		if (targetCurrentHealth < SMH_Configuration.profiles[SMH_Configuration.currentProfile].threshold or UnitInParty('target') or UnitInRaid('target')) then
			-- for a target not in the party, UnitHealth returns the percentage of maximum health
			-- so we must make a guess at how much is gone based on the target's class and level
			if not UnitInParty('target') and not UnitInRaid('target') then
				local slope = (maxClassHealth[UnitClass('target')] - 30) / 59;
				targetMaxHealth = UnitLevel('target') * slope + 30;
				targetCurrentHealth = targetMaxHealth * targetCurrentHealth / 100;
				shouldAddTarget = true;
			elseif (targetCurrentHealth < targetMaxHealth) then
				shouldAddTarget = true;
			end
			if shouldAddTarget then
				local targetPercentHealth = targetCurrentHealth / targetMaxHealth;
				if (targetPercentHealth < (SMH_Configuration.profiles[SMH_Configuration.currentProfile].threshold / 100.0)) then
					table.insert(groupStatus, 1, { id = 'target', percent = targetPercentHealth, current = targetCurrentHealth, max = targetMaxHealth });
					outputDebugMessage("The target "..UnitName('target').." put into heal list");
				end
			end
		end
	end

	--	If ALT is used, the player is put at the front of the list
	if (IsAltKeyDown() == 1) then
		tryToInsertUnitForHealing(groupStatus, 'player', "threshold", "Player put at front of heal list because of ALT");
	end

	return groupStatus;
end

--	This returns nil if there is no spell to cast.
--	The unit passed in is a table.
--	It will check all three types of spell we can cast
--	and rule out ones that cannot be cast based on what
--	the target already has on it, and what the caster
--	is doing/has on it.  It then decides on the spell
--	to cast based on the current algorithm.
local function determineSpellAndRankForUnit(groupInfo, healingMode)
	local playerBuffs = unitBuffs('player');
	local casterHasNaturesSwiftness = playerBuffs[SMH_SPELL_NATURESSWIFTNESS];
	local casterHasClearcasting = playerBuffs[SMH_SPELL_CLEARCASTING];
	local casterHasInnerFocus = playerBuffs[SMH_SPELL_INNER_FOCUS];
	local targetBuffs;
	if (groupInfo.id == 'player') then
		targetBuffs = playerBuffs;
	else
		targetBuffs = unitBuffs(groupInfo.id);
	end
	local targetHasRegrowth = targetBuffs[SMH_SPELL_REGROWTH];
	local targetHasRejuvenation = targetBuffs[SMH_SPELL_REJUVENATION];
	local targetHasRenew = targetBuffs[SMH_SPELL_RENEW];
	local targetHasWeakenedSoul = targetBuffs[SMH_SPELL_WEAKENED_SOUL];
	local targetHasPowerWordShield = targetBuffs[SMH_SPELL_POWER_WORD_SHIELD];
	local rankIndex = { };
	for i = 1, table.getn(classSpellsOfInterest[_SMH_class]), 1 do
		rankIndex[classSpellsOfInterest[_SMH_class][i]] = 1;
	end
	local spellsToCheck = classSpellsOfInterest[_SMH_class];
	local healAmount;
	local spellInfo;
	local trulyDesiredRank;
	local myMana = UnitMana('player');
	local spellName;
	local targetLevel = UnitLevel(groupInfo.id);
	local spellToUse = nil;
	local spellRank = nil;

	outputDebugMessage("Determining spell to use for "..UnitName(groupInfo.id));
	if (nil == myMana) then myMana = 0; end
	for i = 1, table.getn(spellsToCheck), 1 do		-- loop through each of spellsToCheck
		spellName = spellsToCheck[i];

		-- first determine the level of spell needed based on healing required
		while (true) do
			spellInfo = cachedSpellBookData[spellName][rankIndex[spellName]];
			if (spellInfo) then

				--	Determine the base heal amount garnered from the
				--	tooltip for each spell.  This includes modifications
				--	from talents, but not from equipment.  It uses the
				--	technique the user wants for healing to determine
				--	where in the range of healing the target should be.
				healAmount = spellInfo.minHeal + ((spellInfo.maxHeal - spellInfo.minHeal) * SMH_Configuration.profiles[SMH_Configuration.currentProfile].healPercentage / 100);

				--	Now modify the base amount by the amount of extra
				--	healing equipment can do.
				if (_SMH_extraHealingAmount > 0) then
					--	Penalties are integer amounts representing percentages off of _SMH_extraHealingAmount
					local learned;
					local levelPenalty = 0;
					local timePenalty = 0;
					local totalPenalty = 0;

					--	If the spell was learned prior to level 20 we have a penalty of how much of the
					--	_SMH_extraHealingAmount that is applied.
					learned = levelLearned[spellName][rankIndex[spellName]];
					if (learned < 20) then levelPenalty = (20 - learned) * 3.75; else levelPenalty = 0; end

					--	The time penalty for a spell depends on the casting time compared to 3.5 seconds
					--	of if it is Heal Over Time, compared to 15 seconds.  These values are based on the
					--	base time for spells and ignore talent modifications.  Note that Regrowth takes
					--	2 seconds to cast for the initial spell, and the HOT is 21 seconds.  50% is given to
					--	the initial heal and 50% to the HOT.  For Rejuvenation, the HOT takes only 12 seconds
					--	so only 80% of the possible heal is permitted.
					timePenalty = timePenalties[spellName][rankIndex[spellName]];

					totalPenalty = levelPenalty + timePenalty;

--					outputDebugMessage("Initial heal amount: "..healAmount.." Raw Extra Healing: ".._SMH_extraHealingAmount);
					if (totalPenalty > 0) then
						if (totalPenalty < 100) then
							healAmount = healAmount + (_SMH_extraHealingAmount * (100 - totalPenalty) / 100);
						end
					else
						healAmount = healAmount + _SMH_extraHealingAmount;
					end
--					outputDebugMessage("Level Penalty: "..levelPenalty.."% Time Penalty: "..timePenalty.."%");
--					outputDebugMessage("Newly calculated heal amount: "..healAmount);
				end

				if (groupInfo.max - groupInfo.current < healAmount and spellName ~= SMH_SPELL_POWER_WORD_SHIELD) then break end
				rankIndex[spellName] = rankIndex[spellName] + 1;
			else
				rankIndex[spellName] = rankIndex[spellName] - 1;
				break;
			end
		end

		-- now alter the rank based on overhealing
		if spellName ~= SMH_SPELL_POWER_WORD_SHIELD then
			trulyDesiredRank = rankIndex[spellName];
			local amountToAttemptToRaise = SMH_Configuration.profiles[SMH_Configuration.currentProfile].overhealRanks[spellName];
			if amountToAttemptToRaise > 0 then
				while (cachedSpellBookData[spellName][rankIndex[spellName]] and amountToAttemptToRaise > 0) do
					rankIndex[spellName] = rankIndex[spellName] + 1;
					amountToAttemptToRaise = amountToAttemptToRaise - 1;
				end
				if not cachedSpellBookData[spellName][rankIndex[spellName]] and rankIndex[spellName] > 0 then
					rankIndex[spellName] = rankIndex[spellName] - 1;
				end
				if trulyDesiredRank ~= rankIndex[spellName] then
					outputDebugMessage("For "..spellName.." we really wanted rank "..trulyDesiredRank.." but overheal raised that to rank "..rankIndex[spellName]);
				end
			end
		end

		-- now check to ensure we have enough mana to cast the desired spell
		trulyDesiredRank = rankIndex[spellName];
		if not casterHasClearcasting and not casterHasInnerFocus then
			while (rankIndex[spellName] > 0) do
				if (myMana >= cachedSpellBookData[spellName][rankIndex[spellName]].mana) then break end
				rankIndex[spellName] = rankIndex[spellName] - 1;
			end
			if (trulyDesiredRank ~= rankIndex[spellName]) then
				outputDebugMessage("For "..spellName.." decided we want rank "..trulyDesiredRank.." but mana limits us to rank "..rankIndex[spellName]);
			end

		-- now we see if we have Clearcasting and if so raise the rank as high as it can go
		else
			while (cachedSpellBookData[spellName][rankIndex[spellName]]) do
				rankIndex[spellName] = rankIndex[spellName] + 1;
			end
			rankIndex[spellName] = rankIndex[spellName] - 1;	-- get back to where we have a spell
			if (trulyDesiredRank ~= rankIndex[spellName]) then
				outputDebugMessage("For "..spellName.." decided we want rank "..trulyDesiredRank.." but Clearcasting/Inner Focus allows us to use our highest rank "..rankIndex[spellName]);
			end
		end

		-- now check to ensure spell can be cast on level of target
		trulyDesiredRank = rankIndex[spellName];
		if (rankIndex[spellName] > 0) then
			while (rankIndex[spellName] > 0) do
				if (targetLevel + 10 >= cachedSpellTargetLevels[spellName][rankIndex[spellName]]) then break end
				rankIndex[spellName] = rankIndex[spellName] - 1;
			end
			if (trulyDesiredRank ~= rankIndex[spellName]) then
				outputDebugMessage("For "..spellName.." "..UnitName(groupInfo.id).." is only level "..targetLevel.." so had to reduce rank from "..trulyDesiredRank.." to rank "..rankIndex[spellName]);
			end
		end

		outputDebugMessage("For "..spellName.." final rank is "..rankIndex[spellName]);
	end

	-- eliminate spells in cooldown
	if _SMH_class == SMH_CLASS_PRIEST_ENGLISH then
		if not actionNotInCooldown(_SMH_powerWordShieldSlot) then
			rankIndex[SMH_SPELL_POWER_WORD_SHIELD] = 0;
			outputDebugMessage("Cannot cast Power Word: Shield because it is still in cooldown.");
		end
	end

	-- if we are moving and do not have Nature's Swiftness we are limited to instant spells
	if _SMH_isMoving and not casterHasNaturesSwiftness then
		if _SMH_class == SMH_CLASS_DRUID_ENGLISH then
			rankIndex[SMH_SPELL_REGROWTH] = 0;
			rankIndex[SMH_SPELL_HEALING_TOUCH] = 0;
			outputDebugMessage("We are moving without Nature's Swiftness so are limited to Rejuvenation.");
		elseif _SMH_class == SMH_CLASS_PALADIN_ENGLISH then
			rankIndex[SMH_SPELL_HOLY_LIGHT] = 0;
			rankIndex[SMH_SPELL_FLASH_OF_LIGHT] = 0;
			outputDebugMessage("We are moving so cannot heal.");
		elseif _SMH_class == SMH_CLASS_PRIEST_ENGLISH then
			rankIndex[SMH_SPELL_LESSER_HEAL] = 0;
			rankIndex[SMH_SPELL_HEAL] = 0;
			rankIndex[SMH_SPELL_GREATER_HEAL] = 0;
			rankIndex[SMH_SPELL_FLASH_HEAL] = 0;
			outputDebugMessage("We are moving so are limited to Renew and Power Word: Shield.");
		elseif _SMH_class == SMH_CLASS_SHAMAN_ENGLISH then
			rankIndex[SMH_SPELL_HEALING_WAVE] = 0;
			rankIndex[SMH_SPELL_LESSER_HEALING_WAVE] = 0;
			outputDebugMessage("We are moving without Nature's Swiftness so cannot heal.");
		end
	end

	-- eliminate spells the target already has
	if _SMH_class == SMH_CLASS_DRUID_ENGLISH then
		if targetHasRegrowth then
			rankIndex[SMH_SPELL_REGROWTH] = 0;
			outputDebugMessage(UnitName(groupInfo.id).." already has Regrowth");
		end
		if targetHasRejuvenation then
			rankIndex[SMH_SPELL_REJUVENATION] = 0;
			outputDebugMessage(UnitName(groupInfo.id).." already has Rejuvenation");
		end
	elseif _SMH_class == SMH_CLASS_PRIEST_ENGLISH then
		if targetHasRenew then
			rankIndex[SMH_SPELL_RENEW] = 0;
			outputDebugMessage(UnitName(groupInfo.id).." already has Renew");
		end
		if targetHasWeakenedSoul then
			rankIndex[SMH_SPELL_POWER_WORD_SHIELD] = 0;
			outputDebugMessage(UnitName(groupInfo.id).." has Weakened Soul");
		end
		if targetHasPowerWordShield then
			rankIndex[SMH_SPELL_POWER_WORD_SHIELD] = 0;
			outputDebugMessage(UnitName(groupInfo.id).." has Power Word: Shield");
		end
	end

	-- limit ourselves to any type of spell we are forcing ourselves to use
	if healingMode > 0 then
		for i = 1, table.getn(_smh_HealingModeConfiguration[healingMode].nuke), 1 do
			rankIndex[_smh_HealingModeConfiguration[healingMode].nuke[i]] = 0;
		end
		outputDebugMessage("We are limiting ourselves to ".. _smh_HealingModeConfiguration[healingMode].desire .. " because of command.");
	end

	-- now we use our algorithm to determine the spell to cast
	local temp = { };
	for n in pairs(SMH_Configuration.profiles[SMH_Configuration.currentProfile].algorithm) do table.insert(temp, n) end
	table.sort(temp, numericGreaterCompare);
	for i, n in ipairs(temp) do 
		if (groupInfo.percent * 100) > tonumber(n) then
			outputDebugMessage(string.format("the groupinfo percentage is %d and the table shows %s...a match", groupInfo.percent * 100, n));
			if rankIndex[SMH_Configuration.profiles[SMH_Configuration.currentProfile].algorithm[n]] > 0 then
				spellToUse = SMH_Configuration.profiles[SMH_Configuration.currentProfile].algorithm[n];
				outputDebugMessage(string.format("we have found the spell to use in table %s %s", n, spellToUse));
			end
			break;
		end
	end

	-- if no spell has been found using the algorithm do into our default spell list
	if (spellToUse == nil) then
		local tableSize = table.getn(SMH_Configuration.profiles[SMH_Configuration.currentProfile].defaultSpells);
		for i = 1, tableSize do
			if rankIndex[SMH_Configuration.profiles[SMH_Configuration.currentProfile].defaultSpells[i]] > 0 then
				spellToUse = SMH_Configuration.profiles[SMH_Configuration.currentProfile].defaultSpells[i];
				break;
			end
		end
	end

	if (spellToUse) then
		spellRank = rankIndex[spellToUse];
		outputDebugMessage("Decided on "..spellToUse.." Rank("..spellRank..")");
	else
		outputDebugMessage("Found no spell to use");
	end

	return spellToUse, spellRank
end

local function outputPartyMessage(message)
	if SMH_Configuration.profiles[SMH_Configuration.currentProfile].messagesParty then
		if UnitInParty('party1') then
			SendChatMessage(message, "PARTY")
		else
			outputChatMessage(message);
		end
	else
		outputChatMessage(message);
	end
end

--	This function walks through the spell book looking for spells in which we have interest.
--	It then caches the entries into the spell book into tables so we can easily determine
--	which spell to cast based on the name of the spell and the desired casting rank.
function smh_cacheSpellBookData()

	outputDebugMessage("Caching spell book data.");

	-- clear the cache
	cachedSpellBookData = { };
	local localSpellsOfInterest = { };

	for i = 1, table.getn(classSpellsOfInterest[_SMH_class]), 1 do
		cachedSpellBookData[classSpellsOfInterest[_SMH_class][i]] = { };
		localSpellsOfInterest[classSpellsOfInterest[_SMH_class][i]] = true;
	end
	for i = 1, table.getn(classBuffsOfInterest[_SMH_class]), 1 do
		cachedSpellBookData[classBuffsOfInterest[_SMH_class][i]] = { };
		localSpellsOfInterest[classBuffsOfInterest[_SMH_class][i]] = true;
	end
	for i = 1, table.getn(classAttacksOfInterest[_SMH_class]), 1 do
		cachedSpellBookData[classAttacksOfInterest[_SMH_class][i]] = { };
		localSpellsOfInterest[classAttacksOfInterest[_SMH_class][i]] = true;
	end

	local index = 1;

	while (true) do
		local spellName, spellRankName = GetSpellName(index, SpellBookFrame.bookType);
		local maxHealAmount, minHealAmount;

		-- If there is no spellName then we have run out of spells in the spell book
		if (not spellName) then
			do break end
		end

		-- If the spellName is in our table we have interest in processing it
		if (localSpellsOfInterest[spellName]) then
			local _, _, spellRank = string.find(spellRankName, SMH_SPELLBOOK_RANK_FORMAT);
			local rankIndex = tonumber(spellRank);
			--	Now we look at the tooltips for each of the spells to pull out the mana cost
			--	and the amount of healing that can be done.  These values have been adjusted
			--	for the talents already.
			smhTooltip:SetSpell(index, SpellBookFrame.bookType);
			local manaLine = smhTooltipTextLeft2:GetText();
			local _,_,manaCost = string.find(smhTooltipTextLeft2:GetText(), SMH_SPELLBOOK_MANA_FORMAT);
			if spellParsersMaxHeal[_SMH_class][spellName] then
				maxHealAmount = spellParsersMaxHeal[_SMH_class][spellName](smhTooltipTextLeft4:GetText());
			else
				maxHealAmount = 0;
			end
			if spellParsersMinHeal[_SMH_class][spellName] then
				minHealAmount = spellParsersMinHeal[_SMH_class][spellName](smhTooltipTextLeft4:GetText());
			else
				minHealAmount = 0;
			end
			local entry = { id = index, mana = tonumber(manaCost), maxHeal = maxHealAmount, minHeal = minHealAmount };
			cachedSpellBookData[spellName][rankIndex] = entry;
		end

		-- Next spell please
		index = index + 1
	end
end

local _smhSetNamePattern = "^(.*) %(%d/%d%)$";

--	The purpose of this function is to look through all the equipped items
--	and determine how much extra healing is provided by them.
function smh_cacheEquipmentData()

	_SMH_extraHealingAmount = 0;
	local _SMH_setNames = { };
	local canLookForSetPatterns;

	for i, slot in equipmentSlots do
		id, _ = GetInventorySlotInfo(slot.."Slot");
		item, _, _ = smhTooltip:SetInventoryItem('player', id);
		if item then
			canLookForSetPatterns = false;
			local lineCount = smhTooltip:NumLines();
			for j=1, lineCount, 1 do
				local text = getglobal("smhTooltipTextLeft"..j);
				if text then
					local line = text:GetText();

					local match, setName = string.find(line, _smhSetNamePattern);
					if match ~= nil then
						if not _SMH_setNames[setName] then
							canLookForSetPatterns = true;
							_SMH_setNames[setName] = true;
						end
					end

					for k, pattern in _smh_healingPatterns do
						local match, _, amount = string.find(line, pattern);
						if (match ~= nil) then
							outputDebugMessage("On "..slot.." found "..tonumber(amount).." healing.");
							_SMH_extraHealingAmount = _SMH_extraHealingAmount + tonumber(amount);
							break;
						end
					end
					if canLookForSetPatterns then
						for k, pattern in _smh_setHealingPatterns do
							local match, _, amount = string.find(line, pattern);
							if match ~= nil then
								outputDebugMessage("On "..slot.." found "..tonumber(amount).." healing as part of set.");
								_SMH_extraHealingAmount = _SMH_extraHealingAmount + tonumber(amount);
								break;
							end
						end
					end
				end
			end
		end
	end

	outputDebugMessage("Total extra healing found on equipment: ".._SMH_extraHealingAmount);
	com_mithrandir_smhEquipmentHealing:SetText(string.format(SMH_TOTAL_EXTRA_HEALING_FORMAT, _SMH_extraHealingAmount));

end

local function smh_healCommand(healingMode)
	if _SMH_swiftNatureSlot ~= nil and SMH_Configuration.profiles[SMH_Configuration.currentProfile].checkingNaturesSwiftness and not hasBuff('player', SMH_SPELL_NATURESSWIFTNESS, false) and actionNotInCooldown(_SMH_swiftNatureSlot) then
		outputChatMessage(SMH_INFORMATIVE_MESSAGE_CAST_NS);
		UseAction(_SMH_swiftNatureSlot);
		return;
	end
	local groupStatus = orderedHealableUnits();
	if (table.getn(groupStatus) == 0) then
		outputChatMessage(SMH_INFORMATIVE_MESSAGE_NO_ONE);
	else
		if SpellIsTargeting() then SpellStopTargeting() end
		for key, value in groupStatus do
			local spellName, spellRank = determineSpellAndRankForUnit(value, healingMode);
			local shouldBreak = false;
			if (spellName) then
				local spellId = cachedSpellBookData[spellName][spellRank].id;
				if (spellId ~= nil) then
					if (UnitIsVisible(value.id)) then
						if value.id == 'player' then	-- self casting can ignore tons of stuff we do
							outputPartyMessage(string.format(SMH_INFORMATIVE_MESSAGE_HEALINGFORMAT, spellName, spellRank, UnitName(value.id)));
							CastSpell(spellId, SpellBookFrame.bookType);
							SpellTargetUnit('player');
							break;
						else
							local targetBuffs = unitBuffs(value.id);
							if not targetBuffs[SMH_SPELL_PROWL] and not targetBuffs[SMH_SPELL_STEALTH] and not targetBuffs[SMH_SPELL_SHADOWMELD] then

								-- Since we want to find the range we will set the target to the
								-- one we want to heal.  Therefore, we need to know whether we
								-- already had a target.
								local hadTarget = false;
								local targetName = nil;
								if (UnitExists('target')) then
									hadTarget = true;
									if (UnitIsFriend('target', 'player')) then targetName = UnitName('target'); end
								end
								TargetUnit(value.id);

								if (UnitIsUnit('target', value.id)) then
									if (_SMH_healingSlot == nil or IsActionInRange(_SMH_healingSlot) == 1) then
										outputPartyMessage(string.format(SMH_INFORMATIVE_MESSAGE_HEALINGFORMAT, spellName, spellRank, UnitName(value.id)));
										CastSpell(spellId, SpellBookFrame.bookType);
										shouldBreak = true;
									else
										outputChatMessage(string.format(SMH_INFORMATIVE_MESSAGE_OORFORMAT, UnitName(value.id)));
									end
								else
									outputChatMessage(string.format(SMH_INFORMATIVE_MESSAGE_CANNOT_TARGETFORMAT, UnitName(value.id)));
								end

								-- Since we actually set the target to the one we wanted to heal
								-- to be able to determine whether they are in range of our spell
								-- we have to reset the target to what it was before.  Note that
								-- the following line does not work.
								-- if (hadTarget) then TargetLastTarget(); else ClearTarget(); end
								if (hadTarget) then
									if (targetName ~= nil) then TargetByName(targetName); else TargetLastEnemy(); end
								else
									ClearTarget();
								end
								if shouldBreak then break end
							else
								outputChatMessage(string.format(SMH_INFORMATIVE_MESSAGE_STEALTHFORMAT, UnitName(value.id)));
							end
						end
					else
						outputChatMessage(string.format(SMH_INFORMATIVE_MESSAGE_INVISIBLEFORMAT, UnitName(value.id)));
					end
				else
					outputChatMessage(string.format(SMH_INFORMATIVE_MESSAGE_SPELL_LOSTFORMAT, spellName, spellRank, UnitName(value.id)));
				end
				if SpellIsTargeting() then SpellStopTargeting() end
			else
				outputChatMessage(string.format(SMH_INFORMATIVE_MESSAGE_NO_SPELLFORMAT, UnitName(value.id)));
			end
		end
	end
end

local function smh_healHelp()
	local helpCount = table.getn(SMH_HELP_LINES);
	for i = 1, helpCount, 1 do
		outputChatMessage(string.format(SMH_HELP_LINES[i], SMH_MACRO_COMMAND));
	end
end

local outputSettingMessages = {
	["mesagesParty"] = SMH_OUTPUT_MESSAGE_MESSAGE,
	["debug"] = SMH_OUTPUT_MESSAGE_DEBUG,
	["checkingNaturesSwiftness"] = SMH_OUTPUT_MESSAGE_NATURESSWIFTNESS,
	};

local function smh_healBoolean(parameterName, yn)
	SMH_Configuration.profiles[SMH_Configuration.currentProfile][parameterName] = yn;
	outputChatMessage(string.format("<%s>: %s %s", SMH_VERSION_STRING, outputSettingMessages[parameterName], outputBooleanMappings[yn]));
end

local function smhHealNormal()
	smh_healCommand(0);
end

local function smhHealSetProfile(profileName)
	SMH_Configuration.currentProfile = profileName;
	com_mithrandir_smhProfileName:SetText(string.format(SMH_PROFILE_TEXT_FORMAT, SMH_Configuration.currentProfile));
	com_mithrandir_smhMiniButton:SetPoint("TOPLEFT", "Minimap", "TOPLEFT", 52 - (80 * cos(SMH_Configuration.profiles[SMH_Configuration.currentProfile].minimapButtonAngle)), (80 * sin(SMH_Configuration.profiles[SMH_Configuration.currentProfile].minimapButtonAngle)) - 52);

end

--	This routine is only used for testing different things
--	out and has no real purpose other than allowing me to
--	have a test harness available relatively easily.
local function smhHealTest()
end

--	This routine casts Moonfire at the highest level that
--	mana permits unless the target is a totem and then it
--	casts Moonfire at the minimum level needed to destroy
--	the totem.
local function smhMoonfire()
	local casterHasClearcasting = hasBuff('player', SMH_SPELL_CLEARCASTING, false);
	local rankIndex = 1;
	local trulyDesiredRank;
	local myMana = UnitMana('player');
	if (nil == myMana) then myMana = 0; end

	while (true) do
		spellInfo = cachedSpellBookData[SMH_SPELL_MOONFIRE][rankIndex];
		if spellInfo then
			rankIndex = rankIndex + 1;
		else
			rankIndex = rankIndex - 1;
			break;
		end
	end
	outputDebugMessage("Highest rank Moonfire known is "..rankIndex);

	-- Then limit rankIndex if our target is a totem
	local targetName = UnitName('target');
	if targetName == nil then targetName = "NO TARGET NAME" end;
	local totemRank = totems[targetName];
	if totemRank ~= nil then
		rankIndex = totemRank;
		outputDebugMessage("Moonfire rank set to "..rankIndex.." because target is "..targetName);
	end

	-- Now ensure we have the mana to cast it
	trulyDesiredRank = rankIndex;
	if not casterHasClearcasting then
		while (rankIndex > 0) do
			if myMana >= cachedSpellBookData[SMH_SPELL_MOONFIRE][rankIndex].mana then break end
			rankIndex = rankIndex - 1;
		end
		if trulyDesiredRank ~= rankIndex then
			outputDebugMessage("Wanted Moonfire rank "..trulyDesiredRank.." but mana limits us to rank "..rankIndex);
		end
	else
		--	We have Clearcasting so can raise the rank to the highest we have
		while (cachedSpellBookData[SMH_SPELL_MOONFIRE][rankIndex]) do
			rankIndex = rankIndex + 1;
		end
		rankIndex = rankIndex - 1;
		if trulyDesiredRank ~= rankIndex then
			outputDebugMessage("Wanted Moonfire rank "..trulyDesiredRank.." but Clearcasting allows us to rank "..rankIndex);
		end
	end

	local spellId = cachedSpellBookData[SMH_SPELL_MOONFIRE][rankIndex].id;
	outputDebugMessage("Casting Moonfire rank "..rankIndex);
	CastSpell(spellId, SpellBookFrame.bookType);

end

function SMH_ShowHideOptionsUI()
	if (com_mithrandir_smhOptions:IsVisible()) then
		com_mithrandir_smhOptions:Hide();
	else
		com_mithrandir_smhOptions:Show();
	end
end

local commandFunctions = {
	[SMH_COMMAND_HELP] = smh_healHelp,
	[SMH_COMMAND_UI] = SMH_ShowHideOptionsUI,
	[SMH_COMMAND_TEST] = smhHealTest,
	};

local commandParameters = {
	[SMH_COMMAND_MSG] = "messagesParty",
	[SMH_COMMAND_DEBUG] = "debug",
	};

--	This table contains the command-line commands that invoke specific
--	healing spells.  This is populated based on class of user at startup.
local commandHeals = { };

function smhOnSlashCommand(msg)		-- all the parameters will be passed in the msg...therefore, we will need to split some out
	if (msg == nil or msg == "") then
		smhHealNormal();
	else
		msg = string.lower(msg);
		-- hopefully this means we will have command and options filled in such as:
		-- /smh msg => command="msg" options=nil
		-- /smh msg on => command="msg" options="on"
		local _, _, command, options = string.find(msg, "^%s*(%a+)%s*(%a*)");
		if (command ~= nil) then
			if commandHeals[command] then
				smh_healCommand(commandHeals[command]);
			elseif commandFunctions[command] then
				commandFunctions[command]()
			else
				if commandParameters[command] then
					local booleanValue;
					if (options == SMH_COMMAND_OPTION_ON) then booleanValue = true
					elseif (options == SMH_COMMAND_OPTION_OFF) then booleanValue = false
					else booleanValue = not SMH_Configuration.profiles[SMH_Configuration.currentProfile][commandParameters[command]];
					end
					smh_healBoolean(commandParameters[command], booleanValue);
				end
			end
		end
	end
end

local function SMH_Slider_SetTitle(this)
	local id = this:GetID();
	local format = _smh_sliderTitleFormats[id][2];
	getglobal(this:GetName().."Text"):SetText(string.format(format, SMH_Configuration.profiles[SMH_Configuration.currentProfile][_smh_sliderTitleFormats[id][1]]));
end

function SMH_Slider_OnShow()
	local id = this:GetID();
	getglobal(this:GetName().."High"):SetText();
	getglobal(this:GetName().."Low"):SetText();
	SMH_Slider_SetTitle(this);
	if (id == 5) then
		this:SetMinMaxValues(0, 359);
	else
		this:SetMinMaxValues(1, 100);
	end
	this:SetValueStep(1);
	this:SetValue(SMH_Configuration.profiles[SMH_Configuration.currentProfile][_smh_sliderTitleFormats[id][1]]);
end

function SMH_Slider_OnValueChanged()
	local id = this:GetID();
	SMH_Configuration.profiles[SMH_Configuration.currentProfile][_smh_sliderTitleFormats[id][1]] = this:GetValue();
	SMH_Slider_SetTitle(this);
	if (id == 5) then
		com_mithrandir_smhMiniButton:SetPoint("TOPLEFT", "Minimap", "TOPLEFT", 52 - (80 * cos(this:GetValue())), (80 * sin(this:GetValue())) - 52);
	end
end

function SMH_OverhealSlider_SetTitle(this)
	local name = classSpellsOfInterest[_SMH_class][this:GetID()];
	getglobal(this:GetName().."Text"):SetText(string.format("%s: %s", name, SMH_Configuration.profiles[SMH_Configuration.currentProfile].overhealRanks[name]));
end

function SMH_OverhealSlider_OnShow()
	local id = this:GetID();
	local name = classSpellsOfInterest[_SMH_class][id];
	getglobal(this:GetName().."High"):SetText();
	getglobal(this:GetName().."Low"):SetText();
	SMH_OverhealSlider_SetTitle(this);
	this:SetMinMaxValues(0, table.getn(levelLearned[name]) - 1);
	this:SetValueStep(1);
	this:SetValue(SMH_Configuration.profiles[SMH_Configuration.currentProfile].overhealRanks[classSpellsOfInterest[_SMH_class][id]]);
end

function SMH_OverhealSlider_OnValueChanged()
	local id = this:GetID();
	SMH_Configuration.profiles[SMH_Configuration.currentProfile].overhealRanks[classSpellsOfInterest[_SMH_class][id]] = this:GetValue();
	SMH_OverhealSlider_SetTitle(this);
end

function SMH_CheckButton_OnClick()
	local key = _smh_checkButtonTitles[this:GetID()][1];
	SMH_Configuration.profiles[SMH_Configuration.currentProfile][key] = (this:GetChecked() == 1);
end

function SMH_CheckButton_OnLoad()
	getglobal(this:GetName().."Text"):SetText(_smh_checkButtonTitles[this:GetID()][2]);
end

function SMH_CheckButton_OnShow()
	if SMH_Configuration.profiles[SMH_Configuration.currentProfile][_smh_checkButtonTitles[this:GetID()][1]] then
		this:SetChecked(1);
	else
		this:SetChecked(0);
	end
end

local function updateMiniProfileWindow()
	local temp = { };
	local count = 0;

	for n in pairs(SMH_Configuration.profiles) do table.insert(temp, n) end
	table.sort(temp, lesserCompare);
	for i, n in ipairs(temp) do
		if SMH_Configuration.profiles[n]["shouldDisplayProfile"] then
			count = count + 1;
			if SMH_Configuration.currentProfile == n then
				getglobal("com_mithrandir_smhMiniProfilesListItem"..count):SetText("|cFF00FF00"..n.."|r");
			else
				getglobal("com_mithrandir_smhMiniProfilesListItem"..count):SetText(n);
			end
			getglobal("com_mithrandir_smhMiniProfilesListItem"..count):Show();
		end
	end
	for i = count + 1, _SMH_maxNumberInProColumn, 1 do
		getglobal("com_mithrandir_smhMiniProfilesListItem"..i):Hide();
	end
	if count > 1 then
		com_mithrandir_smhMiniProfiles:Show();
	else
		com_mithrandir_smhMiniProfiles:Hide();
	end
end

local function lookupOrderedProfileName(whichOne)
	local retval = nil;
	local temp = { };

	for n in pairs(SMH_Configuration.profiles) do table.insert(temp, n) end
	table.sort(temp, lesserCompare);
	retval = temp[whichOne];
	return retval;
end

function SMH_ProfileCheck_OnClick()
	local profileName = lookupOrderedProfileName(this:GetID());
	SMH_Configuration.profiles[profileName]["shouldDisplayProfile"] = (this:GetChecked() == 1);
	updateMiniProfileWindow();
end

function SMH_ProfileCheck_OnLoad()
	getglobal(this:GetName().."Text"):SetText(SMH_PROFILE_CHECKBUTTON_TITLE);
end

function SMH_ProfileCheck_OnShow()
	local profileName = lookupOrderedProfileName(this:GetID());
	if SMH_Configuration.profiles[profileName]["shouldDisplayProfile"] then
		this:SetChecked(1);
	else
		this:SetChecked(0);
	end
end

function SMH_UpdateClassFrame(frame, number)
	local key = getglobal(frame:GetName().."Title"):GetText();
	getglobal(frame:GetName().."Button0"):SetChecked(0);
	getglobal(frame:GetName().."Button1"):SetChecked(0);
	getglobal(frame:GetName().."Button2"):SetChecked(0);
	getglobal(frame:GetName().."Button3"):SetChecked(0);
	getglobal(frame:GetName().."Button"..number):SetChecked(1);
	SMH_Configuration.profiles[SMH_Configuration.currentProfile].classPriorities[key] = number;
end

function SMH_ClassFrame_OnShow()
	SMH_UpdateClassFrame(this, SMH_Configuration.profiles[SMH_Configuration.currentProfile].classPriorities[getglobal(this:GetName().."Title"):GetText()])
end

--	This routine makes sure only the one clicked is considered checked
--	and makes sure the configuration information is updated with the
--	proper numeric value.
function SMH_ClassRadio_OnClick()
	SMH_UpdateClassFrame(this:GetParent(), this:GetID());
end

--	If forcing it means we overwrite whatever was stored
function SMH_UpdateFromDefaults(forcing)
	if forcing then
		for key in pairs(SMH_Configuration_Defaults) do
			SMH_Configuration.profiles[SMH_Configuration.currentProfile][key] = SMH_Configuration_Defaults[key];
		end
	else
		-- Ensure each profile has the proper keys present
		for profile in pairs(SMH_Configuration.profiles) do
			-- Make sure each default key is present in the profile
			for key in pairs(SMH_Configuration_Defaults) do
				if (nil == SMH_Configuration.profiles[profile][key]) then
					SMH_Configuration.profiles[profile][key] = SMH_Configuration_Defaults[key];
				end
			end
			-- Remove any key from the profile that does not exist in the defaults
			for key in pairs(SMH_Configuration.profiles[profile]) do
				if (nil == SMH_Configuration_Defaults[key]) then
					SMH_Configuration.profiles[profile][key] = nil;
				end
			end
		end
	end
end

local function SMH_UpdateDefaultSpellButtons()
	for i = 1, table.getn(SMH_Configuration.profiles[SMH_Configuration.currentProfile].defaultSpells), 1 do
		getglobal("com_mithrandir_smhAlgControl"..i).tooltiptext = SMH_Configuration.profiles[SMH_Configuration.currentProfile].defaultSpells[i];
		getglobal("com_mithrandir_smhAlgControl"..i):SetNormalTexture(_smh_defaultSpellIcons[SMH_Configuration.profiles[SMH_Configuration.currentProfile].defaultSpells[i]]);
	end
end

function SMH_SwapDefaults(left, right)
	local temp = SMH_Configuration.profiles[SMH_Configuration.currentProfile].defaultSpells[left];
	SMH_Configuration.profiles[SMH_Configuration.currentProfile].defaultSpells[left] = SMH_Configuration.profiles[SMH_Configuration.currentProfile].defaultSpells[right];
	SMH_Configuration.profiles[SMH_Configuration.currentProfile].defaultSpells[right] = temp;
	SMH_UpdateDefaultSpellButtons();
end

local function hideCharacterPriorityButtons()
	for i = 1, _SMH_maxNumberInColumn, 1 do
		getglobal("com_mithrandir_smhCharPriIgnored"..i):Hide();
		getglobal("com_mithrandir_smhCharPriLow"..i):Hide();
		getglobal("com_mithrandir_smhCharPriHigh"..i):Hide();
	end
end

local function hideAlgorithmButtons()
	for i = 1, _SMH_maxNumberInAlgColumn, 1 do
		getglobal("com_mithrandir_smhAlgorithm"..i):Hide();
	end
end

local function hideProfileButtons()
	for i = 1, _SMH_maxNumberInProColumn, 1 do
		getglobal("com_mithrandir_smhProfile"..i):Hide();
		getglobal("com_mithrandir_smhProfileAppear"..i):Hide();
	end
end

local function displayCharacterPriorityButtons()
	local temp = { };

	for n in pairs(SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityIgnored) do table.insert(temp, n) end
	table.sort(temp, lesserCompare);
	for i, n in ipairs(temp) do
		getglobal("com_mithrandir_smhCharPriIgnored"..i):SetText(n);
		getglobal("com_mithrandir_smhCharPriIgnored"..i):Show();
	end
	for i = table.getn(temp) + 1, _SMH_maxNumberInColumn, 1 do
		getglobal("com_mithrandir_smhCharPriIgnored"..i):Hide();
	end

	temp = { };
	for n in pairs(SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityLow) do table.insert(temp, n) end
	table.sort(temp, lesserCompare);
	for i, n in ipairs(temp) do
		getglobal("com_mithrandir_smhCharPriLow"..i):SetText(n);
		getglobal("com_mithrandir_smhCharPriLow"..i):Show();
	end
	for i = table.getn(temp) + 1, _SMH_maxNumberInColumn, 1 do
		getglobal("com_mithrandir_smhCharPriLow"..i):Hide();
	end

	temp = { };
	for n in pairs(SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityHigh) do table.insert(temp, n) end
	table.sort(temp, lesserCompare);
	for i, n in ipairs(temp) do
		getglobal("com_mithrandir_smhCharPriHigh"..i):SetText(n);
		getglobal("com_mithrandir_smhCharPriHigh"..i):Show();
	end
	for i = table.getn(temp) + 1, _SMH_maxNumberInColumn, 1 do
		getglobal("com_mithrandir_smhCharPriHigh"..i):Hide();
	end
end

local function displayAlgorithmButtons()
	local temp = { };

	for n in pairs(SMH_Configuration.profiles[SMH_Configuration.currentProfile].algorithm) do table.insert(temp, n) end
	table.sort(temp, numericGreaterCompare);
	for i, n in ipairs(temp) do
		getglobal("com_mithrandir_smhAlgorithm"..i):SetText(string.format("%s%% %s", n, SMH_Configuration.profiles[SMH_Configuration.currentProfile].algorithm[n]));
		getglobal("com_mithrandir_smhAlgorithm"..i):Show();
	end
	for i = table.getn(temp) + 1, _SMH_maxNumberInAlgColumn, 1 do
		getglobal("com_mithrandir_smhAlgorithm"..i):Hide();
	end
end

local function displayProfileButtons()
	local temp = { };

	for n in pairs(SMH_Configuration.profiles) do table.insert(temp, n) end
	table.sort(temp, lesserCompare);
	for i, n in ipairs(temp) do
		getglobal("com_mithrandir_smhProfile"..i):SetText(n);
		getglobal("com_mithrandir_smhProfile"..i):Show();
		getglobal("com_mithrandir_smhProfileAppear"..i):Show();
	end
	for i = table.getn(temp) + 1, _SMH_maxNumberInProColumn, 1 do
		getglobal("com_mithrandir_smhProfile"..i):Hide();
		getglobal("com_mithrandir_smhProfileAppear"..i):Hide();
	end
end

function SMH_CharacterPriorityAdd()
	if UnitIsHealable('target') then
		local targetName = UnitName('target');
		if SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityHigh[targetName] then
			SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityHigh[targetName] = nil;
		end
		if SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityLow[targetName] then
			SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityLow[targetName] = nil;
		end
		if SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityIgnored[targetName] then
			SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityIgnored[targetName] = nil;
		end
		local id = this:GetID();
		if id == 0 then
			if table.getn(SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityIgnored) < _SMH_maxNumberInColumn then
				SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityIgnored[targetName] = true;
			end
		elseif id == 1 then
			if table.getn(SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityLow) < _SMH_maxNumberInColumn then
				SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityLow[targetName] = true;
			end
		elseif id == 2 then
			if table.getn(SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityHigh) < _SMH_maxNumberInColumn then
				SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityHigh[targetName] = true;
			end
		end
		displayCharacterPriorityButtons();
	end
end

function SMH_CharacterPriorityRemove()
	local id = this:GetID();
	local characterName = this:GetText();

	if id < 100 then
		SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityIgnored[characterName] = nil;
	elseif id < 200 then
		SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityLow[characterName] = nil;
	else
		SMH_Configuration.profiles[SMH_Configuration.currentProfile].charactersPriorityHigh[characterName] = nil;
	end
	displayCharacterPriorityButtons();
end

function SMH_AlgorithmRemove()
	local buttonTitle = this:GetText();
	local match, _, amount = string.find(buttonTitle, "^(%d+)%%.*");
	if (match ~= nil) then
		SMH_Configuration.profiles[SMH_Configuration.currentProfile].algorithm[amount] = nil;
		displayAlgorithmButtons();
	end
end

function SMH_AlgorithmControl()
	local temp = { };

	for n in pairs(SMH_Configuration.profiles[SMH_Configuration.currentProfile].algorithm) do table.insert(temp, n) end
	if table.getn(temp) < _SMH_maxNumberInAlgColumn then
		_SMH_algorithmAddingSpell = this.tooltiptext;
		com_mithrandir_smhAlgorithmAcceptInfo:SetText(string.format(SMH_ALGORITHM_INSTRUCTIONS, _SMH_algorithmAddingSpell));
		com_mithrandir_smhAlgorithmAccept:Show();
	end
end

function SMH_AlgorithmAdd()
	local percentageValue = com_mithrandir_smhAlgorithmAcceptField:GetText();

	if tonumber(percentageValue) >= 0 and tonumber(percentageValue) <= 99 then
		SMH_Configuration.profiles[SMH_Configuration.currentProfile].algorithm[percentageValue] = _SMH_algorithmAddingSpell;
		displayAlgorithmButtons();
	end
end

function SMH_ProfileSet()
	smhHealSetProfile(this:GetText());
	updateMiniProfileWindow();
end

local function getOrderMiniProfileName(whichOne)
	local temp = { };

	for n in pairs(SMH_Configuration.profiles) do
		if SMH_Configuration.profiles[n]["shouldDisplayProfile"] then
			table.insert(temp, n)
		end
	end
	table.sort(temp, lesserCompare);
	return temp[whichOne]
end

function SMH_MiniProfileSet()
	local profileName = getOrderMiniProfileName(this:GetID());
	smhHealSetProfile(profileName);
	updateMiniProfileWindow();
	if com_mithrandir_smhOptions:IsVisible() then
		SMH_SetTab(_SMH_currentTab);
	end
end

local function smhProfileWithName(aLowercaseName)
	local retval = nil;
	local count = 0;

	for n in pairs(SMH_Configuration.profiles) do
		if (string.lower(n) == aLowercaseName) then
			retval = SMH_Configuration.profiles[n];
		end
		count = count + 1;
	end
	return retval, count;
end

local function smhCreateNewProfileFromCurrent(newProfileName)
	local temp = { };

	for n in pairs(SMH_Configuration.profiles[SMH_Configuration.currentProfile]) do
		temp[n] = SMH_Configuration.profiles[SMH_Configuration.currentProfile][n];
	end
	SMH_Configuration.profiles[newProfileName] = temp;
	end

function SMH_NewProfile()
	local newProfileName = com_mithrandir_smhProfileAcceptField:GetText();

	if newProfileName ~= '' then
		local existing, count = smhProfileWithName(string.lower(newProfileName));
		if existing == nil and count < _SMH_maxNumberInProColumn then
			smhCreateNewProfileFromCurrent(newProfileName);
			smhHealSetProfile(newProfileName);
			displayProfileButtons();
			updateMiniProfileWindow();
		end
	end
end

local function smhFirstProfile()
	local retval = nil;

	for n in pairs(SMH_Configuration.profiles) do
		if retval == nil then
			retval = n;
		end
	end;
	return retval;
end

function SMH_DeleteProfile()
	local existing, count = smhProfileWithName(string.lower(SMH_Configuration.currentProfile));

	if existing and count > 1 then
		SMH_Configuration.profiles[SMH_Configuration.currentProfile] = nil;
		smhHealSetProfile(smhFirstProfile());
		displayProfileButtons();
		updateMiniProfileWindow();
	end
end

function SMH_RenameProfile()
	local newProfileName = com_mithrandir_smhProfileRenameField:GetText();

	if newProfileName ~= '' then
		local existing, count = smhProfileWithName(string.lower(newProfileName));
		if existing == nil then
			smhCreateNewProfileFromCurrent(newProfileName);
			SMH_Configuration.profiles[SMH_Configuration.currentProfile] = nil;
			smhHealSetProfile(newProfileName);
			displayProfileButtons();
			updateMiniProfileWindow();
		end
	end
end

local function smh_init()
	outputChatMessage(string.format(SMH_IntroductionTextFormat, SMH_VERSION_STRING, SMH_MACRO_COMMAND));

	SlashCmdList["SMARTHEAL"] = function(msg)
		smhOnSlashCommand(msg)
	end
	SLASH_SMARTHEAL1 = SMH_MACRO_COMMAND;

	--	Ensure our tooltip is not messed up.
	if not smhTooltip:IsOwned(UIParent) then
		smhTooltip:SetOwner(UIParent, "ANCHOR_NONE");
	end

	--	Populate the spells of interest convenience table.
	local tableSize = table.getn(classSpellsOfInterest[_SMH_class]);
	for i = 1, tableSize, 1 do
		spellsOfInterest[classSpellsOfInterest[_SMH_class][i]] = true;
	end

	--	Populate the help system with the class specific lines
	for i = 1, table.getn(SMH_CLASS_HELP_LINES[_SMH_class]), 1 do
		table.insert(SMH_HELP_LINES, i + 1, SMH_CLASS_HELP_LINES[_SMH_class][i]);
	end

	--	Here we populate the default configuration information
	--	with data that is class specific.
	if _SMH_class == SMH_CLASS_DRUID_ENGLISH then
		SMH_Configuration_Defaults.algorithm = {
			["75"] = SMH_SPELL_REJUVENATION,
			["50"] = SMH_SPELL_REGROWTH,
			["25"] = SMH_SPELL_HEALING_TOUCH,
			["0"] = SMH_SPELL_REGROWTH,
			};
		SMH_Configuration_Defaults.defaultSpells = { 
			SMH_SPELL_REJUVENATION,
			SMH_SPELL_REGROWTH,
			SMH_SPELL_HEALING_TOUCH,
			};
		SMH_Configuration_Defaults.overhealRanks = {
			[SMH_SPELL_REJUVENATION] = 0,
			[SMH_SPELL_REGROWTH] = 0,
			[SMH_SPELL_HEALING_TOUCH] = 0,
			};
		commandParameters[SMH_COMMAND_NATURE] = "checkingNaturesSwiftness";
		commandFunctions[SMH_COMMAND_MOONFIRE] = smhMoonfire;
		commandHeals[SMH_COMMAND_HEALINGTOUCH] = 1;
		commandHeals[SMH_COMMAND_REGROWTH] = 2;
		commandHeals[SMH_COMMAND_REJUVENATION] = 3;
	elseif _SMH_class == SMH_CLASS_PALADIN_ENGLISH then
		SMH_Configuration_Defaults.algorithm = {
			["75"] = SMH_SPELL_HOLY_LIGHT,
			["0"] = SMH_SPELL_FLASH_OF_LIGHT,
			};
		SMH_Configuration_Defaults.defaultSpells = { 
			SMH_SPELL_FLASH_OF_LIGHT,
			SMH_SPELL_HOLY_LIGHT,
			};
		SMH_Configuration_Defaults.overhealRanks = {
			[SMH_SPELL_FLASH_OF_LIGHT] = 0,
			[SMH_SPELL_HOLY_LIGHT] = 0,
			};
		commandHeals[SMH_COMMAND_HOLYLIGHT] = 4;
		commandHeals[SMH_COMMAND_FLASHOFLIGHT] = 5;
	elseif _SMH_class == SMH_CLASS_PRIEST_ENGLISH then
		SMH_Configuration_Defaults.algorithm = {
			["75"] = SMH_SPELL_FLASH_HEAL,
			["0"] = SMH_SPELL_GREATER_HEAL,
			};
		SMH_Configuration_Defaults.defaultSpells = {
			SMH_SPELL_GREATER_HEAL,
			SMH_SPELL_HEAL,
			SMH_SPELL_LESSER_HEAL,
			SMH_SPELL_RENEW,
			SMH_SPELL_FLASH_HEAL,
			SMH_SPELL_POWER_WORD_SHIELD,
			};
		SMH_Configuration_Defaults.overhealRanks = {
			[SMH_SPELL_GREATER_HEAL] = 0,
			[SMH_SPELL_HEAL] = 0,
			[SMH_SPELL_LESSER_HEAL] = 0,
			[SMH_SPELL_RENEW] = 0,
			[SMH_SPELL_FLASH_HEAL] = 0,
			};
		commandHeals[SMH_COMMAND_LESSERHEAL] = 6;
		commandHeals[SMH_COMMAND_HEAL] = 7;
		commandHeals[SMH_COMMAND_GREATERHEAL] = 8;
		commandHeals[SMH_COMMAND_RENEW] = 9;
		commandHeals[SMH_COMMAND_FLASHHEAL] = 10;
		commandHeals[SMH_COMMAND_POWERWORDSHIELD] = 13;
	elseif _SMH_class == SMH_CLASS_SHAMAN_ENGLISH then
		SMH_Configuration_Defaults.algorithm = {
			["75"] = SMH_SPELL_LESSER_HEALING_WAVE,
			["0"] = SMH_SPELL_HEALING_WAVE,
			};
		SMH_Configuration_Defaults.defaultSpells = { 
			SMH_SPELL_HEALING_WAVE,
			SMH_SPELL_LESSER_HEALING_WAVE,
			};
		SMH_Configuration_Defaults.overhealRanks = {
			[SMH_SPELL_HEALING_WAVE] = 0,
			[SMH_SPELL_LESSER_HEALING_WAVE] = 0,
			};
		commandHeals[SMH_COMMAND_HEALINGWAVE] = 11;
		commandHeals[SMH_COMMAND_LESSERHEALINGWAVE] = 12;
	end

	--	In version 0.9.7 we have added profiles and therefore we need to detect whether the
	--	SMH_Configuration that is read in has the proper profiles set up and if not we take
	--	the stored values and convert them to a default profile.
	if nil == SMH_Configuration.currentProfile then
		local temp = { };
		for key in pairs(SMH_Configuration) do
			temp[key] = SMH_Configuration[key];
			SMH_Configuration[key] = nil;
		end
		SMH_Configuration.currentProfile = SMH_DEFAULT_PROFILE_NAME;
		SMH_Configuration.profiles = { };
		SMH_Configuration.profiles[SMH_Configuration.currentProfile] = temp;
	end

	SMH_UpdateFromDefaults(false);

	--	If the current profile does not actually exist in the set of profiles
	--	we set the current profile to something that does exist.
	if nil == SMH_Configuration.profiles[SMH_Configuration.currentProfile] then
		smhHealSetProfile(smhFirstProfile());
	end

	--	In version 0.9.7 we have added the capability to do Power Word: Shield and this means
	--	every priest using earlier versions will not have it in their default spells so we add
	--	it here.
	if _SMH_class == SMH_CLASS_PRIEST_ENGLISH then
		if table.getn(SMH_Configuration.profiles[SMH_Configuration.currentProfile].defaultSpells) < 6 then
			table.insert(SMH_Configuration.profiles[SMH_Configuration.currentProfile].defaultSpells, SMH_SPELL_POWER_WORD_SHIELD);
		end
	end
	
	com_mithrandir_smhMiniButton:SetPoint("TOPLEFT", "Minimap", "TOPLEFT", 52 - (80 * cos(SMH_Configuration.profiles[SMH_Configuration.currentProfile].minimapButtonAngle)), (80 * sin(SMH_Configuration.profiles[SMH_Configuration.currentProfile].minimapButtonAngle)) - 52);

	com_mithrandir_smhOptionsTitle:SetText("SMart Heal "..SMH_VERSION.." Options");

	com_mithrandir_smhProfileName:SetText(string.format(SMH_PROFILE_TEXT_FORMAT, SMH_Configuration.currentProfile));

	SMH_UpdateDefaultSpellButtons();

	updateMiniProfileWindow();

	this:RegisterEvent("PLAYER_ENTERING_WORLD");
	this:RegisterEvent("PLAYER_LEAVING_WORLD");
end

function smh_onLoad()
	_, _SMH_class = UnitClass('player');
	_SMH_myName = UnitName('player');
	if classesPermittedToUseAddon[_SMH_class] then
		this:RegisterEvent("VARIABLES_LOADED");
	else
		com_mithrandir_smhMiniButton:Hide();
	end
end

local function SMH_FrameUpdate()
	local numberTabs = getn(_smh_optionTabs);
	for i = 1, numberTabs, 1 do
		getglobal("com_mithrandir_smhTab" .. i):Show();
	end
end

local uiElements = { };

function SMH_SetTab(tabId)
	local reshowing = false;
	if tabId == _SMH_currentTab then reshowing = true end;
	_SMH_currentTab = tabId;
	if (tabId == 8) then
		SMH_priorityString();
	end
	if (tabId == 4) then
		displayCharacterPriorityButtons();
	else
		hideCharacterPriorityButtons();
	end
	if (tabId == 3) then
		displayAlgorithmButtons();
		SMH_UpdateDefaultSpellButtons();
	else
		hideAlgorithmButtons();
	end
	if (tabId == 7) then
		displayProfileButtons();
	else
		hideProfileButtons();
	end
	local numberUIElements = getn(uiElements);
	for i = 1, numberUIElements, 1 do
		if tabId == uiElements[i][2] then
			if reshowing then uiElements[i][1]:Hide() end;
			uiElements[i][1]:Show();
		else
			uiElements[i][1]:Hide();
		end
	end
end

function SMH_Tab_OnClick(tabId)
	if not tabId then
		tabId = this:GetID();
	end
	SMH_SetTab(tabId);
	local numberTabs = getn(_smh_optionTabs);
	for i = 1, numberTabs, 1 do
		if (i ~= tabId) then
			getglobal("com_mithrandir_smhTab" .. i):SetChecked(nil);
		else
			getglobal("com_mithrandir_smhTab" .. i):SetChecked(1);
		end
	end
	SMH_FrameUpdate();
end

function com_mithrandir_smhOptionsOnLoad()
	if classesPermittedToUseAddon[_SMH_class] then
		-- Currently we set up the uiElements by hand because they are not
		-- loaded until just before we get this call.
		uiElements = {
			{ com_mithrandir_smhMessagesParty, 1 },
			{ com_mithrandir_smhUsesTarget, 2 },
			{ com_mithrandir_smhHealsParty, 2 },
			{ com_mithrandir_smhHealsPets, 2 },
			{ com_mithrandir_smhHealingRaid, 2 },
			{ com_mithrandir_smhHealingRaidPets, 2 },
			{ com_mithrandir_smhPartyPriority, 2 },
			{ com_mithrandir_smhDamageThreshold, 2 },
			{ com_mithrandir_smhPetDamageThreshold, 2 },
			{ com_mithrandir_smhSelfDamageThreshold, 2 },
			{ com_mithrandir_smhHealPercentage, 2 },
			{ com_mithrandir_smhDebug, 1 },
			{ com_mithrandir_smhMiniButtonAngle, 1 },
			{ com_mithrandir_smhReset, 1 },
			{ com_mithrandir_smhRaidInstruct, 5 },
			{ com_mithrandir_smhPriRaid1, 5 },
			{ com_mithrandir_smhPriRaid2, 5 },
			{ com_mithrandir_smhPriRaid3, 5 },
			{ com_mithrandir_smhPriRaid4, 5 },
			{ com_mithrandir_smhPriRaid5, 5 },
			{ com_mithrandir_smhPriRaid6, 5 },
			{ com_mithrandir_smhPriRaid7, 5 },
			{ com_mithrandir_smhPriRaid8, 5 },
			{ com_mithrandir_smhPriRaidPets, 2 },
			{ com_mithrandir_smhDefaultInstruct, 3 },
			{ com_mithrandir_smhPriorityDesc, 8 },
			{ com_mithrandir_smhClassPriority, 5 },
			{ com_mithrandir_smhClassHighPriority, 2 },
			{ com_mithrandir_smhOverhealing, 6 },
			{ com_mithrandir_smhPriorityDruid, 5 },
			{ com_mithrandir_smhPriorityHunter, 5 },
			{ com_mithrandir_smhPriorityMage, 5 },
			{ com_mithrandir_smhPriorityPaladin, 5 },
			{ com_mithrandir_smhPriorityPriest, 5 },
			{ com_mithrandir_smhPriorityRogue, 5 },
			{ com_mithrandir_smhPriorityShaman, 5 },
			{ com_mithrandir_smhPriorityWarlock, 5 },
			{ com_mithrandir_smhPriorityWarrior, 5 },
			{ com_mithrandir_smhOverheal1, 6 },
			{ com_mithrandir_smhOverheal2, 6 },
			{ com_mithrandir_smhHealsText, 2 },
			{ com_mithrandir_smhPrioritiesText, 2 },
			{ com_mithrandir_smhPriorityText, 4 },
			{ com_mithrandir_smhCharacterPriorityLow, 4 },
			{ com_mithrandir_smhCharacterPriorityHigh, 4 },
			{ com_mithrandir_smhCharacterPriorityIgnored, 4 },
			{ com_mithrandir_smhAlgControl1, 3 },
			{ com_mithrandir_smhAlgControl2, 3 },
			{ com_mithrandir_smhDefaultLeft2, 3 },
			{ com_mithrandir_smhAlgorithm, 3 },
			{ com_mithrandir_smhProfile, 7 },
			{ com_mithrandir_smhNewProfile, 7 },
			{ com_mithrandir_smhDeleteProfile, 7 },
			{ com_mithrandir_smhRenameProfile, 7 },
			};

		if _SMH_class == SMH_CLASS_DRUID_ENGLISH then
			table.insert(uiElements, { com_mithrandir_smhNaturesSwiftness, 1 });
			table.insert(uiElements, { com_mithrandir_smhOverheal3, 6 });
			table.insert(uiElements, { com_mithrandir_smhAlgControl3, 3 });
			table.insert(uiElements, { com_mithrandir_smhDefaultLeft3, 3 });
		elseif _SMH_class == SMH_CLASS_PRIEST_ENGLISH then
			table.insert(uiElements, { com_mithrandir_smhOverheal3, 6 });
			table.insert(uiElements, { com_mithrandir_smhOverheal4, 6 });
			table.insert(uiElements, { com_mithrandir_smhOverheal5, 6 });
			table.insert(uiElements, { com_mithrandir_smhAlgControl3, 3 });
			table.insert(uiElements, { com_mithrandir_smhAlgControl4, 3 });
			table.insert(uiElements, { com_mithrandir_smhAlgControl5, 3 });
			table.insert(uiElements, { com_mithrandir_smhAlgControl6, 3 });
			table.insert(uiElements, { com_mithrandir_smhDefaultLeft3, 3 });
			table.insert(uiElements, { com_mithrandir_smhDefaultLeft4, 3 });
			table.insert(uiElements, { com_mithrandir_smhDefaultLeft5, 3 });
			table.insert(uiElements, { com_mithrandir_smhDefaultLeft6, 3 });
		end

		SMH_SetTab(1);
		com_mithrandir_smhTab1:SetChecked(1);
		local numberTabs = getn(_smh_optionTabs);
		for i = 1, numberTabs, 1 do
			getglobal("com_mithrandir_smhTab" .. i).tooltiptext = _smh_optionTabs[i][1];
			getglobal("com_mithrandir_smhTab" .. i):SetNormalTexture(_smh_optionTabs[i][2]);
			getglobal("com_mithrandir_smhTabButton" .. i).tooltiptext = _smh_optionTabs[i][1];
		end

		SMH_FrameUpdate();
	end
end

_SMH_event_dispatch['VARIABLES_LOADED'] = function()
	smh_init();
end

_SMH_event_dispatch['SPELLS_CHANGED'] = function(arg)
	if (not arg) then	-- arg is nil when new spell learned
		smh_cacheSpellBookData();
	end
end

_SMH_event_dispatch['PLAYER_ENTERING_WORLD'] = function()
	smh_cacheSpellBookData();
	findHealingSlot();
	smh_cacheEquipmentData();
	this:RegisterEvent("ACTIONBAR_SLOT_CHANGED");
	this:RegisterEvent("UNIT_INVENTORY_CHANGED");
	this:RegisterEvent("CHARACTER_POINTS_CHANGED");
	this:RegisterEvent("LEARNED_SPELL_IN_TAB");
end

_SMH_event_dispatch['PLAYER_LEAVING_WORLD'] = function()
	this:UnregisterEvent("ACTIONBAR_SLOT_CHANGED");
	this:UnregisterEvent("UNIT_INVENTORY_CHANGED");
	this:UnregisterEvent("CHARACTER_POINTS_CHANGED");
	this:UnregisterEvent("LEARNED_SPELL_IN_TAB");
end

_SMH_event_dispatch['LEARNED_SPELL_IN_TAB'] = function()
	smh_cacheSpellBookData();
end

_SMH_event_dispatch['ACTIONBAR_SLOT_CHANGED'] = function(arg1)
	--	arg1 tells us the slot that has changed
	if (_SMH_healingSlot == nil or _SMH_healingSlot == arg1
		or (_SMH_swiftNatureSlot == nil and _SMH_class == SMH_CLASS_DRUID_ENGLISH) or _SMH_swiftNatureSlot == arg1
		or (_SMH_powerWordShieldSlot == nil and _SMH_class == SMH_CLASS_PRIEST_ENGLISH) or _SMH_powerWordShieldSlot == arg1
		) then
		findHealingSlot();
	end
end

_SMH_event_dispatch['UNIT_INVENTORY_CHANGED'] = function(arg1)
	--	arg1 tells us what unit had the inventory change
	if (arg1 == 'player') then
		smh_cacheEquipmentData();
	end
end

_SMH_event_dispatch['CHARACTER_POINTS_CHANGED'] = function(arg1)
	--	arg1 tells us the delta in talent points
	if (arg1 == -1) then
		smh_cacheSpellBookData();	-- because spell tooltips are modified based on talents
	end
end

function smh_onEvent()
	if _SMH_event_dispatch[event] then
		_SMH_event_dispatch[event](arg1, arg2)
	end
end

function smh_onUpdate(delay)
	if classesPermittedToUseAddon[_SMH_class] then
		local positionX, positionY = GetPlayerMapPosition('player');
		if positionX ~= _SMH_lastX or positionY ~= _SMH_lastY then
			_SMH_isMoving = true;
		else
			_SMH_isMoving = false;
		end
		_SMH_lastX = positionX;
		_SMH_lastY = positionY;
	end
end
